'''
Bliss based File Transfer Implementation
'''
import urlparse

import errno
import sys
import os
import stat
import logging
import traceback
import time

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../.."))
#sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
from pilot.api import State
from bigjob import logger
import bliss.saga as saga

class BlissFileAdaptor(object):
    """ BigData File Management for Pilot Data """
    
   
    
    def __init__(self, service_url):        
        self.service_url = service_url
        result = urlparse.urlparse(service_url)
        self.host = result.netloc
        self.query = result.path      
       
        self.__state=State.New
                        
        
    def get_security_context(self):
        """ Returns security context that needs to be available on the distributed
            node in order to access this Pilot Data """
        return None
    
    
    def initialize_pilotdata(self):
        # check whether directory exists

        self.__state=State.Running
        
        
    def get_pilotdata_size(self):
        # check size
        return None
    
    
    def delete_pilotdata(self):
        self.__state=State.Done
    
        
    def get_state(self):
        return self.__state
        
            
    def create_du(self, du_id):
        du_dir = os.path.join(self.path, str(du_id))
        logger.debug("mkdir: " + du_dir)
        
        
        
    def put_du(self, du):
        pass       
                
    
    def copy_du_to_url(self, du,  local_url, remote_url):
        pass

        

    def copy_du(self, du, pd_new):
        remote_url = pd_new.service_url + "/" + str(du.id)
        local_url =  self.service_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
    
    def get_du(self, du, target_url):
        remote_url = target_url
        local_url =  self.service_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
        
        
    def remove_du(self, du):
        self.__remove_directory(os.path.join(self.path, du.id))
    
    
    ###########################################################################
    # Pure File Management APIs
        
    def transfer(self, source_url, target_url):
        pass
    
    def create_remote_directory(self, target_url):
        return True
    
    
    def get_path(self, target_url):
        result = urlparse.urlparse(target_url)
        target_query = result.path
        
        
    ###########################################################################
    
   
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    
    
if __name__ == "__main__":
    go = GlobusOnlineFileAdaptor("http://drelu:bigjob@cli.globusonline.org?ep=drelu#egi&path=/ho")
    go.transfer("go://cli.globusonline.org?ep=drelu#MacBook&path=/~/cert.tar.gz", "go://cli.globusonline.org?ep=xsede#kraken&path=/~/")
    


#!/usr/bin/env python

from bigjob import logger
import os
import uuid
import time
import traceback
import sys
from boto.ec2.connection import EC2Connection
from boto.ec2.regioninfo import RegionInfo
import boto.ec2

import bliss.saga as saga
import bliss

###############################################################################
# EC2 General
PLACEMENT_GROUP=None
SECURITY_GROUP="default"

# VM/Image specific configurations
# Recommendation Ubuntu > 1104
# apt-get install gcc python-all-dev git subversion vim
# EC2_AMI_ID="ami-c7943cae" #  ami-82fa58eb official Amazon Ubuntu 12.04 LTS (requires dev tools installation)
# EC2_AMI_ID="ami-d7f742be"
# EC2_USERNAME="ubuntu"
# EC2_KEYNAME="lsu-keypair"
# EC2_KEYNAME="MyKey"

# Authentication
# Please use ~/.boto file to configure your security credentials (if possible)
# see http://boto.readthedocs.org/en/latest/boto_config_tut.html
# 
# [Credentials]
# aws_access_key_id = <your access key>
# aws_secret_access_key = <your secret key>
#
# Alternatively you can use these two variables
AWS_ACCESS_KEY_ID=None
AWS_SECRET_ACCESS_KEY=None




class State:
    UNKNOWN="unknown"
    PENDING="pending"
    RUNNING="running"
    
    
class Service(object):
    """ Plugin for Amazon EC2 and EUCA
    
        Manages endpoint in the form of:
        
            ec2+ssh://<EC2 Endpoint>
            euca+ssh://<EUCA Endpoint>
    """

    def __init__(self, resource_url, pilot_compute_description=None):
        """Constructor"""
        self.resource_url = resource_url
        self.pilot_compute_description = pilot_compute_description
            
    
    def create_job(self, job_description):
        j = Job(job_description, self.resource_url, self.pilot_compute_description)
        return j
            
    
    def __del__(self):
        pass
    
    
    

class Job(object):
    """ Plugin for Amazon EC2 
    
        Starts VM and executes BJ agent on this VM    
        
        
        Eucalyptus on FutureGrid uses a self-signed certificate, which 1) needs to be added to boto configuration
        or 2) certificate validation needs to be disabled.
    """

    def __init__(self, job_description, resource_url, pilot_compute_description):
        
        self.job_description = job_description
        logger.debug("URL: " + str(resource_url) + " Type: " + str(type(resource_url)))
        self.resource_url = saga.Url(str(resource_url))
        self.pilot_compute_description = pilot_compute_description
        
        self.id="bigjob-" + str(uuid.uuid1())
        self.network_ip=None
        
        self.ec2_conn=None
        
        if self.resource_url.scheme == "euca+ssh" or self.resource_url.scheme == "nova+ssh":
            host = self.resource_url.host
            path = "/services/Eucalyptus"
            if self.resource_url.path!=None: 
                path = self.resource_url.path
            port = 8773
            if self.resource_url.port != None:
                port = self.resource_url.port
            region = None
            logger.debug("Host: %s, Path: %s, Port: %d"%(host, path, port))
            if self.resource_url.scheme == "euca+ssh":
                region = RegionInfo(name="eucalyptus", endpoint=host)
            elif self.resource_url.scheme == "nova+ssh":
                region = RegionInfo(name="openstack", endpoint=host)
            logger.debug("Access Key: %s Secret: %s"%(self.pilot_compute_description["access_key_id"],
                                                      self.pilot_compute_description["secret_access_key"]))
            self.ec2_conn = EC2Connection(aws_access_key_id=self.pilot_compute_description["access_key_id"],
                                          aws_secret_access_key=self.pilot_compute_description["secret_access_key"], 
                                          region=region,
                                          is_secure=False,
                                          port=port,
                                          path=path)
        else:
            aws_region = None
            if self.pilot_compute_description.has_key("region"):
                region =  self.pilot_compute_description["region"]
                logger.debug("Connect to region: %s" + region)
                aws_region = boto.ec2.get_region(region,  
                                                 aws_access_key_id=self.pilot_compute_description["access_key_id"], 
                                                 aws_secret_access_key=self.pilot_compute_description["secret_access_key"]
                                                 )

                
            self.ec2_conn = EC2Connection(aws_access_key_id=self.pilot_compute_description["access_key_id"], 
                                          aws_secret_access_key=self.pilot_compute_description["secret_access_key"],
                                          region = aws_region)
            
        self.instance = None
        
        
    def run(self):
        """ Start VM and start BJ agent via SSH on VM """
        
        """ Map fields of Pilot description to EC2 API
            { "vm_id":"ami-d7f742be",
              "vm_ssh_username":"ubuntu",
              "vm_ssh_keyname":"MyKey",
              "vm_ssh_keyfile":"<path>",
              "vm_type":"t1.micro",
              "access_key_id":"xxx",
              "secret_access_key":"xxx"
            }
        """    
            
        reservation = self.ec2_conn.run_instances(self.pilot_compute_description["vm_id"],
                                    key_name=self.pilot_compute_description["vm_ssh_keyname"],
                                    instance_type=self.pilot_compute_description["vm_type"],
                                    security_groups=[SECURITY_GROUP])
                
        self.instance = reservation.instances[0]
        self.instance_id = self.instance.id
        logger.debug("Started EC2/Eucalyptus/Nova instance: %s"%self.instance_id)
        time.sleep(5)
        self.wait_for_running()
        
        if self.resource_url.scheme != "euca+ssh" and self.resource_url.scheme != "nova+ssh":
            self.ec2_conn.create_tags([self.instance_id], {"Name": self.id})
      
        
        self.network_ip = self.instance.ip_address 
        url = "ssh://" + str(self.network_ip)
        logger.debug("Connect to: %s"%(url))
        js = saga.job.Service(url)
        
        # Submit job
        ctx = saga.Context()
        ctx.type = saga.Context.SSH
        ctx.userid = self.pilot_compute_description["vm_ssh_username"]
        ctx.userkey = self.pilot_compute_description["vm_ssh_keyfile"]
        js.session.contexts = [ctx]

        logger.debug("Job Description Type: " + str(type(self.job_description)))

        job = js.create_job(self.job_description)
        
        TRIAL_MAX=30
        trials=0
        while trials < TRIAL_MAX:
            try:
                logger.debug("Attempt: %d, submit pilot job to: %s "%(trials,str(url)))
                job.run()
                break
            except:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                logger.warning("Submission failed: " + str(exc_value))
                #self.__print_traceback()
                trials = trials + 1 
                time.sleep(30)
                if trials == TRIAL_MAX:
                    raise Exception("Submission of agent failed.") 
                
        logger.debug("Job State : %s" % (job.get_state())) 
        
        

    def wait_for_running(self):
        while self.get_state()!=State.RUNNING:
            time.sleep(5)
        
    
    def get_state(self):
        result = State.UNKNOWN
        try:
            self.instance.update()
            result=self.instance.state
        except:
            logger.warning("Instance not reachable/active yet...")
        return result
    
    
    def cancel(self):
        self.instance.terminate()
        
        
    ###########################################################################
    # private methods
    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stdout)
    
 
     
if __name__ == "__main__":
    ec2_service = Service("ec2+ssh://aws.amazon.com")
    j = ec2_service.create_job("blas")
    j.run()
    print j.get_state()



""" Pilot Compute related entities."""

class State(object):
    """ B{State}.
    
        States used for L{PilotCompute}, L{PilotData}, L{ComputeUnit}, L{DataUnit} and L{ComputeDataService}
    """
    Unknown = "Unknown"
    New = "New"
    Running = "Running"
    Done = "Done"
    Canceled = "Canceled"
    Failed = "Failed"
    Pending = "Pending"
    


class PilotComputeDescription(dict):
    """ B{PilotComputeDescription (PCD).}
        
        A PilotComputeDescription is a based on the attributes defined on 
        the SAGA Job Description.

        The PilotComputeDescription is used by the application to specify 
        what kind of PilotJobs it requires.
        
        Example::
             pilot_compute_description = {
                             "service_url": 'fork://localhost',
                             "number_of_processes": 1,                             
                             "working_directory": "/tmp/pilot-compute/",
                             'affinity_datacenter_label': "eu-de-south",              
                             'affinity_machine_label': "mymachine-1" 
                            }
        
        B{Attention}: The PilotComputeDescription is mapped 1:1 to the underlying SAGA-Python (Bliss) 
        job description, which is used for launching the pilot. Depending on the resource, it is required
        to add additional attributes, e.g. some XSEDE/Torque resources require the specification
        of both number_of_process and processes_per_node:
        
        Example::
              pilot_compute_description = {
                             "service_url": 'pbs+ssh://india.futuregrid.org',
                             "number_of_processes": 8,
                             "processes_per_node":8,                     
                             "working_directory": "/N/u/luckow",
                             'affinity_datacenter_label': "us-east-indiana",              
                             'affinity_machine_label': "india" 
                            }
    
        
         
    """

    # Class members
    __slots__ = (
        # Pilot / Agent description
        'service_url',
        'project',
        'working_directory',
        # I/O
        'input',
        'error',
        'output',
        'file_transfer',
        
        # Parallelism
        'number_of_processes',      # Total number of processes to start
        'processes_per_host',       # Nr of processes per host
        
        # Requirements
        'wall_time_limit',
        'queue',
        
        # Affinity
        'affinity_datacenter_label',    # pilot jobs sharing the same label are located in the same data center          
        'affinity_machine_label',       # pilot jobs sharing the same label are located on the same machine
    )

   
    def __init__(self):
        self.service_url.__doc__="SAGA-Python URL for respective resource manager, e.g. fork://localhost"

    
    
    def __setattr__(self, attr, value):
        self[attr]=value
        
    
    def __getattr__(self, attr):
        return self[attr]
    
    

class PilotCompute(object):
    """ B{PilotCompute (PC)}. 
    
        This is the object that is returned by the PilotComputeService when a 
        new PilotCompute (aka Pilot-Job) is created based on a PilotComputeDescription.

        The PilotCompute object can be used by the application to keep track 
        of PilotComputes that are active.
        
        A PilotCompute has state, can be queried, can be cancelled and be 
        re-initialized.
    """

    # Class members
    __slots__ = (
        'id',           # Reference to this PJ
        'description',  # Description of PilotCompute
        'context',      # SAGA context
        'rm',           # Resource Manager URL
        'pj_type',      # Type of TROY backend
        'state',        # State of the PilotCompute
        'state_detail', # Adaptor specific state of the PilotCompute
        'callback',     # Callback object
        'wall_time_left'      # Remaining wallclock time left
    )
    
    def __init__(self):
        raise  NotImplementedError("Abstract super class, please use PilotCompute implementation class in pilot namespace")

    def cancel(self):        
        """ Remove the PilotCompute from the PilotCompute Service.

            Keyword arguments:
            None
        """
        pass

    def reinitialize(self, pilotjob_description):        
        """ Re-Initialize the PilotCompute to the (new) PilotComputeDescription.
        
            Keyword arguments:
            pilotjob_description -- A PilotComputeDescription
        """
        pass

    def set_callback(self, member, cb):
        """ Set a callback function for a member.

            Keyword arguments:
            member -- The member to set the callback for (state / state_detail).
            cb -- The callback object to call.
        """
        pass

    def unset_callback(self, member):
        """ Unset a callback function from a member

            Keyword arguments:
            member -- The member to unset the callback from.
        """
        pass
    
    def get_state(self):
        pass
#
#  Pilot ComputeService
#
class PilotComputeService(object):
    """  B{PilotComputeService (PCS).}
    
        The PilotComputeService is responsible for creating and managing 
        the PilotComputes.

        It is the application's interface to the Pilot-Manager in the 
        P* Model.
        
    """

    #Class members
    __slots__ = (
        'id',           # Reference to this PJS
        'state',       # Status of the PJS
        'pilot_jobs'    # List of PJs under this PJS
    )
    
    

    def __init__(self, pjs_id=None):
        """ Create a PilotComputeService object.

            Keyword arguments:
            pjs_id -- Don't create a new, but connect to an existing (optional)
        """
        raise  NotImplementedError("Abstract super class, please use PilotComputeService implementation class in pilot namespace")


    def create_pilot(self, rm, pilotcompute_description, pj_type=None, context=None):
        """ Add a PilotCompute to the PilotComputeService

            Keyword arguments:
            rm -- Contact string for the resource manager
            pilotcompute_description -- PilotCompute Description
            pj_type --  backend type (optional)
            context -- Security context (optional)

            Return value:
            A PilotCompute handle
        """
        pass

    def cancel(self):
        """ Cancel the PilotComputeService.

            This also cancels all the PilotJobs that were under control of this PJS.

            Keyword arguments:
            None

            Return value:
            Result of operation
        """
        pass


#
#  ComputeUnitService
# 
class ComputeUnitService(object):
    """ B{ComputeUnitService (CUS).}
    
        Please use ComputeDataService!
        
        The ComputeUnitService is the application's interface to submit 
        ComputeUnits to the Pilot-Manager in the P* Model.

        It can provide the application with a list of ComputeUnits that are 
        managed by the Pilot-Manager.

        The ComputeUnitService is linked to a PilotComputeService for the actual 
        execution of the ComputeUnits.
               
    """

    def __init__(self, wus_id=None):
        """ Create a Work Service object.
    
            Keyword arguments:
            wus_id -- Reconnect to an existing WUS (optional).
        """
        raise  NotImplementedError("Abstract super class, please use ComputeUnitService implementation class in pilot namespace")



    def add_pilot_job_service(self, pjs):
        """ Add a PilotComputeService to this WUS.

            Keyword arguments:
            pilotjob_services -- The PilotCompute Service(s) to which this 
                                 Work Unit Service will connect.

            Return:
            Result
        """
        pass


    def remove_pilot_job_service(self, pjs):
        """ Remove a PilotComputeService from this WUS.

            Note that it won't cancel the PilotComputeService, it will just no
            longer be connected to this WUS.

            Keyword arguments:
            pilotjob_services -- The PilotCompute Service(s) to remove from this
                                 Work Unit Service. 

            Return:
            Result
        """
        pass


    def submit_compute_unit(self, wud):
        """ Submit a WU to this Work Unit Service.

            Keyword argument:
            wud -- The WorkUnitDescription from the application

            Return:
            WorkUnit object
        """
        pass


    def cancel(self):
        """ Cancel the WUS.
            
            Cancelling the WUS also cancels all the WUs submitted to it.
    
            Keyword arguments:
            None
    
            Return:
            Result
        """
        pass
    
    def get_state(self):
        pass
    
    def get_id(self):
        pass



#
#  ComuteUnitDescription
# 
class ComputeUnitDescription(dict):
    """ B{ComputeUnitDescription (CUD).}
    
        The ComputeUnitDescription is a task description based on 
        SAGA Job Description. 
        
        It offers the application to describe a L{ComputeUnit} in an abstract 
        way that is dealt with by the Pilot-Manager. It can contain
        references to depended L{DataUnit}s. ComputeUnitDescription are submitted
        to the L{ComputeDataService}.
        
        Format::
        
            compute_unit_description =            
                {
                    'executable': <path to executable>,  
                    'arguments': <arguments>,           # Arguments 
                    'environment': <environment>,          # environment variables 
                    
                    # Working directory
                    # A pilot data url will be translated to the local pilot store url, e.g.
                    #     pilotdata://localhost/c2fafeae-03a9-11e1-9934-109addae22a3' 
                    #     =>           
                    #     ssh://localhost/tmp/pilotdata/c2fafeae-03a9-11e1-9934-109addae22a3
                    #
                    #     i.e. job is executed in /tmp/pilotdata/c2fafeae-03a9-11e1-9934-109addae22a3
                    #     where file can be accessed using relative paths
                    'working_directory': <working directory>,  
                    
                    # I/O
                    'input': <stdin>,
                    'error': <sterr>,
                    'output': <stdout>,
                                
                    # Parallelism
                    'number_of_processes': <Total number of processes to start>,
                    'processes_per_host':  <Nr of processes per host>,
                    'threads_per_process': <Nr of threads to start per process>,
                    'total_core_count':    <Total number of cores requested>,
                    'spmd_variation':      <Type and startup mechanism>,
                    
                    # Requirements
                    'candidate_hosts': [<url>, ...],
                    'cpu_architecture': <architecture, e.g. x86_64>,
                    'total_physical_memory': <memory, e.g. 2000>,
                    'operating_system_type': <os, e.g. Linux>,
                    'total_cpu_time': <cpu time>,
                    'wall_time_limit': <walltime in sec, e.g. 600 (sec)>,            
                    
                    # Data - input/output data flow for ComputeUnit
                    'input_data': [<data unit url>, ... ],      
                    'output_data': [<data unit url>, ... ]
                }
            
            Example::
                    compute_unit_description = {
                            "executable": "/bin/cat",
                            "arguments": ["test.txt"],
                            "number_of_processes": 1,
                            "output": "stdout.txt",
                            "error": "stderr.txt",   
                            "input_data" : [data_unit.get_url()], # this stages the content of the data unit to the working directory of the compute unit
                            "affinity_datacenter_label": "eu-de-south",              
                            "affinity_machine_label": "mymachine-1" 
                        }     
        ComputeUnitDescription objects are loosely typed. A dictionary containing the respective keys
        can be passed instead to the L{ComputeDataService}.
    """
    def __init__(self):
        pass


    def __setattr__(self, attr, value):
        self[attr]=value
        
    
    def __getattr__(self, attr):
        return self[attr]


#
#  ComputeUnit(WU)
# 
class ComputeUnit(object):
    """ B{ComputeUnit (CU).}
    
        This is the object that is returned by the ComputeUnitService when a 
        new ComputeUnit is created based on a ComputeUnitDescription.

        The ComputeUnit object can be used by the application to keep track 
        of ComputeUnits that are active.

        A ComputeUnit has state, can be queried and can be cancelled.
    """

    # Class members
    __slots__ = (
        'id',               # Reference to this WU
        'description',      # Description of this WU
        'state_detail',     # Detailed (application specific) state of this WU
        'callback'          # Callback object
    )

    
    def cancel(self):
        """ Cancel the CU. """
        raise  NotImplementedError("Abstract super class, please use ComputeUnit implementation class in pilot namespace")


    
    def set_callback(self, member, cb):
        """ Set a callback function for a member.

            Keyword arguments:
            member -- The member to set the callback for (state / state_detail).
            cb -- The callback object to call.
        """
        pass

    
    def unset_callback(self, member):
        """ Unset a callback function from a member

            Keyword arguments:
            member -- The member to unset the callback from.
        """
        pass
    

    def get_state(self):
        pass
    
    
    def wait(self):
        """ Wait until in Done state 
            (or Failed state)
        """
#
#  Callback (Abstract) Class
#
class Callback(object):
    """ Callback class.

        Specifies the structure for callback classes.

        Callbacks can be set for WorkUnits on the state or state_detail members.
    """

    def cb(self, wu, member, value):
        """ This is the method that needs to be implemented by the application
        
            Keyword arguments:
            wu -- The WU that is calling back.
            member -- The member that triggered the callback.
            value -- The new (detailed) state.

            Return:
            Keep -- Keep or remove the callback
        """
        pass



import sys
import os
import time
import logging
logging.basicConfig(level=logging.DEBUG)

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot import PilotComputeService, PilotDataService, ComputeDataService, State

COORDINATION_URL = "redis://localhost:6379"

if __name__ == "__main__":        
    
    # What files? Create Pilot Data Description using absolute URLs
    base_dir = "/Users/luckow/workspace-saga/applications/pilot-store/test/data1"
    url_list = os.listdir(base_dir)
    # make absolute paths
    absolute_url_list = [os.path.join(base_dir, i) for i in url_list]
    data_unit_description1 = {
                               "file_urls":absolute_url_list,
                               'affinity_datacenter_label': "eu-de-south",              
                               'affinity_machine_label': "mymachine-1" 
                               }
    logging.debug("Pilot Data Description 1: \n%s"%str(data_unit_description1))
    
    
    # What files? Create Pilot Data Description using remote SSH URLs
    # make remotete paths
    remote_url_list = ["ssh://localhost"+os.path.join(base_dir, i) for i in url_list]
    data_unit_description2 = {
                               "file_urls":remote_url_list,
                               'affinity_datacenter_label': "eu-de-south",              
                               'affinity_machine_label': "mymachine-2" 
                              }
    
    logging.debug("Pilot Data Description 2: \n%s"%str(data_unit_description2))
        
    
    
    # create pilot data service (factory for pilot stores (physical, distributed storage))
    pilot_data_service = PilotDataService(coordination_url=COORDINATION_URL)
    ps1 = pilot_data_service.create_pilot({
                                'service_url': "ssh://localhost/tmp/pilotdata-1/",
                                'size':100,
                               'affinity_datacenter_label': "eu-de-south",              
                               'affinity_machine_label': "mymachine-1"
                                })
    
    ps2 = pilot_data_service.create_pilot({
                                'service_url': "ssh://localhost/tmp/pilotdata-2/",
                                'size':100,
                               'affinity_datacenter_label': "eu-de-south",              
                               'affinity_machine_label': "mymachine-2"
                                })
    
    
      # create pilot data service
    compute_data_service = ComputeDataService()
  
    # add resources to pilot data service    
    compute_data_service.add_pilot_data_service(pilot_data_service) 
    
    ###########################################################################
    # DU1 should only be scheduled to machine 1
    # DU2 should only be scheduled to machine 2
    du1 = compute_data_service.submit_data_unit(data_unit_description1)
    du2 = compute_data_service.submit_data_unit(data_unit_description2)
    
    logging.debug("Finished setup of Pilot Data and Compute Data Service. Waiting for scheduling of Data Units")
    
    compute_data_service.wait()
    logging.debug("DU scheduled: " + du1.url)
    logging.debug("Export files of PD")
    du1.export("/tmp/pilot-data-export/")
    #du2.export("ssh://hotel.futuregrid.org/N/u/luckow/pilot-store-export/")
        
    print "***************************************************************"
    print "To reconnect to Data Unit 1 use the following URL: %s"%du1.url
    print "Run:\n\n " + sys.executable + " example-pilot-data-reconnect.py %s"%du1.url
    print "\n\n******************* SLEEPING *********************************"
    #time.sleep(1200)
    
    logging.debug("Terminate Pilot Data/Compute Data Service")
    compute_data_service.cancel()
    pilot_data_service.cancel()



""" B{PilotCompute Module}: Implementation of L{PilotComputeService}, L{PilotCompute}, L{ComputeUnit}.
"""

import sys
import getopt
import time
import pdb
import os
import traceback
import logging
import uuid


""" import bigjob classes """
from bigjob import logger
from bigjob import bigjob, subjob, description

""" import API objects """
from pilot.api import PilotCompute, PilotComputeService, ComputeUnit, State

""" Configure coordination backend here """
#from pilot.coordination.advert import AdvertCoordinationAdaptor as CoordinationAdaptor
from pilot.coordination.redis_adaptor import RedisCoordinationAdaptor as CoordinationAdaptor


""" This variable defines the coordination system that is used by BigJob
    e.g. 
        advert://localhost (SAGA/Advert SQLITE)
        advert://advert.cct.lsu.edu:8080 (SAGA/Advert POSTGRESQL)
        redis://localhost:6379 (Redis at localhost)
        tcp://localhost (ZMQ)
"""
COORDINATION_URL = "redis://localhost"


class PilotComputeService(PilotComputeService):
    """ B{PilotComputeService (PCS).}
    
        Factory for L{PilotCompute}s.. 
                    
    """
    PJS_ID_PREFIX="pcs-"   


    def __init__(self, coordination_url=COORDINATION_URL, pcs_url=None):
        """ Create a PilotJobService object.

            Keyword arguments:
            pcs_id -- Don't create a new, but connect to an existing (optional)
        """
        self.pilot_computes=[]
        self.coordination_url=coordination_url
        self.coordination_queue=""
        if pcs_url==None:      # new pcs          
            self.id = self.PJS_ID_PREFIX+str(uuid.uuid1())
            self.url = os.path.join(self.coordination_url, "pcs", self.id)
            self.coordination_queue = "PilotComputeServiceQueue-" + str(self.id)
            logger.debug("Created Pilot Compute Service: %s"%self.url)
        else:
            logger.error("Reconnect to PilotComputeService currently not supported.")
        
            

    def create_pilot(self, rm=None, pilot_compute_description=None, pj_type=None, context=None):
        """ Add a PilotJob to the PilotJobService

            Keyword arguments:
            pilot_compute_description -- PilotJob Description
            
            Return value:
            A PilotCompute object
        """
        bj_dict = self.__translate_pj_bj_description(pilot_compute_description)
        bj = self.__start_bigjob(bj_dict)
        pj = PilotCompute(self, bj, pilot_compute_description)
        self.pilot_computes.append(pj)
        return pj
           
    
    def list_pilots(self):
        """ List managed L{PilotCompute}s.

            Return value:
            A list of L{PilotCompute} urls

        """
        return self.pilot_computes
        
    
    def cancel(self):
        """ Cancel the PilotComputeService.

            This also cancels all the PilotJobs that were under control of this PJS.

            Keyword arguments:
            None

            Return value:
            Result of operation
        """
        for i in self.pilot_computes:
            i.cancel()
            
    
    def __repr__(self):
        status_string = "{\n"
        for i in self.pilot_computes:
            status_string = status_string + " PJ: %s, State: %s;"%(i.get_url(), i.get_state())
        status_string = status_string + "\n}"
        return status_string
    
    
    ###########################################################################
    # Internal methods
    def __translate_pj_bj_description(self, pilot_compute_description):
        resource_description={}
        if pilot_compute_description.has_key("service_url"):
            resource_description["resource_url"] = pilot_compute_description["service_url"] 
            
        if pilot_compute_description.has_key("queue"):
            resource_description["queue"] = pilot_compute_description["queue"] 
        else:
            resource_description["queue"] = None
            
        if pilot_compute_description.has_key("allocation"):
            resource_description["allocation"] = pilot_compute_description["allocation"] 
        else:
            resource_description["allocation"] = None
        
        for i in pilot_compute_description.keys():
            resource_description[i] = pilot_compute_description[i] 
        
        resource_description["pilot_compute_description"] = pilot_compute_description
        
        return resource_description

    
    def __start_bigjob(self, bj_dict):
        """ private method - starts a bigjob on the defined resource """
        gram_url = bj_dict["resource_url"]
        logger.debug("start bigjob at: " + gram_url)
        bj = bigjob(self.coordination_url)
        ppn="1"
        if ("processes_per_node" in bj_dict):
            ppn=bj_dict["processes_per_node"]
        else:
            bj_dict["processes_per_node"]="1"

        walltime = 3600
        if ("walltime" in bj_dict):
            walltime=bj_dict["walltime"]

        working_directory = None
        if ("working_directory" in bj_dict):
            working_directory=bj_dict["working_directory"]
            
        bj_filetransfer = None
        if ("file_transfer" in bj_dict):
            bj_filetransfer = bj_dict["file_transfer"]


        bj.start_pilot_job(lrms_url = gram_url,
                           number_nodes = bj_dict["number_of_processes"],
                           queue = bj_dict["queue"],
                           project = bj_dict["allocation"],
                           working_directory = working_directory, 
                           walltime = walltime,
                           processes_per_node = ppn,
                           filetransfers = bj_filetransfer,
                           external_queue = self.coordination_queue,
                           pilot_compute_description = bj_dict["pilot_compute_description"]
                           )
        return bj
    
###############################################################################
    
class PilotCompute(PilotCompute):
    """ B{Pilot Compute (PC).}
    
        This is the object that is returned by the PilotComputeService when a 
        new PilotCompute is created based on a PilotComputeDescription.

        The PilotCompute object can be used by the application to keep track 
        of active pilots.
        
        A ComputePilot has state, can be queried and cancelled.

        Properties:

          - state:            
            The state of the pilot.
         
          - id:
            The id may be 'None' if the Pilot is not yet in Running state.
            The returned ID can be used to connect to the CP instance later 
            on, for example from within a different application instance.  
            type: string (url)

          - pilot_compute_description:
            The PilotComputeDescription used to create this pilot.            
    
    """     
    def __init__(self, pilot_compute_service=None, 
                       bigjob_object=None, 
                       pilot_compute_description=None,
                       pilot_url=None): # for reconnecting
        
        """ Create/reconnect to a Pilot Compute.  

            Keyword arguments:
            pilot_url   -- restore from cp_id

            The implementation will attempt to reconnect to the PC instance
            referenced by the pilot_url.  

        """        
        self.__subjobs = []
        self.__pilot_compute_service = None
        if pilot_url==None:
            logger.debug("Create PilotCompute for BigJob: " + str(bigjob_object))
            self.pilot_compute_description=pilot_compute_description
            self.__pilot_compute_service=pilot_compute_service
            self.__bigjob = bigjob_object        
        else:
            logger.debug("Reconnect to an existing Pilot Compute")
            self.__bigjob = bigjob(pilot_url=pilot_url)
        
        # Store the URL of pilot compute service for later reference
        # This URL is used as central queue for a set of BJs in the
        # ComputeDataServiceDecentral
        if self.__pilot_compute_service!=None:
            self.coordination_queue = pilot_compute_service.coordination_queue
            
        
    def cancel(self):
        """ Terminates the pilot """
        self.__bigjob.cancel()    
    
    
    def get_state(self):
        """ Returns the state of the pilot """
        return self.__bigjob.get_state()    
    
    
    def wait(self):
        """ Wait until Pilot Compute to enter a final state (Done, Cancel or Failed) 

        It is not an error to call wait() in a final state -- the call simply
        returns immediately.        
        """
        self.__bigjob.wait()
    
    
    def list_compute_units(self):
        """ list managed L{ComputeUnit}s.

            Return value:
            A list of L{ComputeUnit} IDs

            The returned list can include units which have been submitted to
            this pilot.
        """
        sj_list = self.__bigjob.list_subjobs()
        cu_list = []
        for i in sj_list:
            cu_list.append(ComputeUnit(cu_url=i))
        return cu_list
    
    
    def get_url(self):
        """ Get unique URL referencing the Pilot Compute 
            This URL can be used to reconnect to the Pilot Compute        
        """
        return self.__bigjob.get_url()
    
    
    def get_free_nodes(self):
        """ Returns the number of free slots available within the pilot """
        return self.__bigjob.get_free_nodes()
    
    
    def get_details(self):
        """ returns a dict that contains the details of the Pilot Compute, 
                - job state
                - description
                - ...        
        """
        return self.__bigjob.get_details()
       
    
    def submit_compute_unit(self, compute_unit_description):
        """
            Submit a CU to this pilot.

            @param compute_unit_description: The L{ComputeUnitDescription} or dictionary describing
                                             the compute task

            @return: L{ComputeUnit} object

            The CUD is (possibly translated and) passed on to the PDS scheduler,
            which will attempt to instantiate the described workload process on
            the managed set of Pilot Computes.  
            
            On success, the returned CU is in Pending state (or moved into any
            state downstream from Pending).

            The call will will honor all attributes set on the CUD.  Attributes which
            are not explicitly set are interpreted as having default values (see
            documentation of CUD), or, where default values are not specified,
            are ignored.
        """
        cu = ComputeUnit(compute_unit_description)
        return self._submit_cu(cu)
    
    
    def __repr__(self):
        return str(self.__bigjob)
        
        
    ###########################################################################
    # Internal methods
        
    def _submit_cu(self, compute_unit):
        """ Submits compute unit to Bigjob """
        logger.debug("Submit CU to big-job")
        sj = subjob()
        sj.submit_job(self.__bigjob.pilot_url, compute_unit.subjob_description)
        self.__subjobs.append(sj)
        compute_unit._update_subjob(sj)
        return compute_unit
        

###############################################################################
        
class ComputeUnit(ComputeUnit):
    """ B{ComputeUnit (CU).}
    
        This is the object that is returned by the ComputeDataService when a 
        new ComputeUnit is submitted based on a ComputeUnitDescription.

        The ComputeUnit object can be used by the application to keep track 
        of ComputeUnits that are active.

        A ComputeUnit has state, can be queried and can be cancelled. 
        
    """
    
    CU_ID_PREFIX="cu-"  

    def __init__(self, compute_unit_description=None, compute_data_service=None, cu_url=None):
        
        if cu_url==None:
            self.id = self.CU_ID_PREFIX + str(uuid.uuid1())
            if compute_data_service!=None:
                self.url = compute_data_service.url + "/" + self.id
                logger.debug("Created CU: %s"%self.url)  
            self.state = State.New       
            self.__subjob = None # reference to BigJob Subjob 
            self.compute_unit_description = compute_unit_description # CU Description
            self.subjob_description = self.__translate_cu_sj_description(compute_unit_description)
        else:
            self.__subjob = subjob(subjob_url=cu_url)
           
    
    def get_id(self):
        """Returns unique identifier of Compute Unit. Deprecated: Please use get_url() instead."""
        return self.id
    
    
    def get_url(self):   
        """Returns URL of Compute Unit. This URL can be used to reconnect to this Compute Unit later on."""
        if self.__subjob!=None:      
            return self.__subjob.get_url()
        else:
            return self.get_id()   
        
    
    def get_details(self):
        """Returns dict with Compute Unit Details (e.g. job description, timings)"""   
        if self.__subjob!=None:      
            return self.__subjob.get_details()
        else:
            return None      
    
    
    def get_state(self):
        """Returns current state of Compute Unit"""
        if self.__subjob != None:
            self.state = self.__subjob.get_state()
        return self.state
    
    
    def wait(self):
        """ Wait until in Done state 
            (or Failed state)
        """
        while True:
            state = self.get_state()
            logger.debug("Compute Unit: %s, State: %s"%(self.id, state))            
            if state==State.Done or state==State.Failed:
                break
            time.sleep(2)

    
    def cancel(self):
        """Terminates Compute Unit"""
        if self.__subjob != None:
            return self.__subjob.cancel()
        return None
    
    def __repr__(self):
        return self.id

    
    def _update_compute_unit_description(self, compute_unit_description):
        self.compute_unit_description = compute_unit_description # CU Description
        self.subjob_description = self.__translate_cu_sj_description(compute_unit_description)

    def _update_subjob(self, subjob):
        self.__subjob = subjob
        
    # INTERNAL
    def __translate_cu_sj_description(self, compute_unit_description):
        jd = description()
        if compute_unit_description.has_key("executable"): 
            jd.executable = compute_unit_description["executable"]
        if compute_unit_description.has_key("spmd_variation"):
            jd.spmd_variation = compute_unit_description["spmd_variation"]
        else:
            jd.spmd_variation = "single"
        if compute_unit_description.has_key("arguments"): 
            jd.arguments = compute_unit_description["arguments"]
        if compute_unit_description.has_key("environment"):
            jd.environment = compute_unit_description["environment"] 
        
        # handling number of processes
        if compute_unit_description.has_key("number_of_processes"):
            jd.number_of_processes=int(compute_unit_description["number_of_processes"])
        elif compute_unit_description.has_key("total_cpu_count"):
            jd.number_of_processes=int(compute_unit_description["total_cpu_count"])
        else:
            jd.number_of_processes=1
            
        if compute_unit_description.has_key("working_directory"): 
            jd.working_directory = compute_unit_description["working_directory"]
        if compute_unit_description.has_key("output"): 
            jd.output =  compute_unit_description["output"]
        if compute_unit_description.has_key("error"): 
            jd.error = compute_unit_description["error"]
        if compute_unit_description.has_key("file_transfer"):
            jd.file_transfer=compute_unit_description["file_transfer"]  
        if compute_unit_description.has_key("input_data"):
            jd.input_data=compute_unit_description["input_data"]  
        if compute_unit_description.has_key("output_data"):
            jd.output_data=compute_unit_description["output_data"]            
        return jd
        
        



""" B{ComputeDataServiceDecentral Module}: A decentral implementation of a ComputeDataService (see L{ComputeDataServiceDecentral}).

A Meta-Scheduling service for pilots (both PilotCompute and PilotData)
"""

import sys
import os
import time
import threading
import logging
import pdb
import Queue
import uuid
import traceback
import urlparse

import bigjob
from bigjob import logger, bigjob, subjob, description

import pilot
from pilot.api import ComputeDataService, State
from pilot.api.api import PilotError
from pilot.impl.pilotdata_manager import PilotData, DataUnit
from pilot.impl.pilotcompute_manager import PilotCompute, PilotComputeService
from pilot.impl.pilot_manager import ComputeUnit


#from pilot.coordination.advert import AdvertCoordinationAdaptor as CoordinationAdaptor
from pilot.coordination.redis_adaptor import RedisCoordinationAdaptor as CoordinationAdaptor

""" Loaded Module determines scheduler:
    
    bigdata.scheduler.data_compute_scheduler - selects random locations for PD and CUs
    bigdata.scheduler.data_compute_affinity_scheduler - considers affinity descriptions
    
"""
from pilot.scheduler.data_compute_affinity_scheduler import Scheduler

class ComputeDataServiceDecentral(ComputeDataService):
    """ B{ComputeDataServiceDecentral.}
    
        The ComputeDataService is the application's interface to submit 
        ComputeUnits and PilotData/DataUnit to the Pilot-Manager 
        in the P* Model.
        
        The decentral ComputeDateService will only work with Redis!
    """    
    CDS_ID_PREFIX="cds-"  

    def __init__(self, cds_url=None):
        """ Create a ComputeDataService (Decentral) object.

            @param cds_url: Reconnect to an existing CDS (optional).
        """
        # Pilot Data
        self.data_units={}
        self.pilot_data_services=[]
        
        # Pilot Compute
        self.compute_units={}
        self.pilot_job_services=[]
            
        if cds_url == None:
            self.id=self.CDS_ID_PREFIX + str(uuid.uuid1())
            application_url = CoordinationAdaptor.get_base_url(pilot.application_id)
            self.url = CoordinationAdaptor.add_cds(application_url, self)            
        else:
            self.id = self.__get_cds_id(cds_url)
            self.url = cds_url
           
        # Background Thread for scheduling
        self.scheduler = Scheduler()
        self.du_queue = Queue.Queue()
        
        self.stop=threading.Event()
        self.scheduler_thread=threading.Thread(target=self._scheduler_thread)
        self.scheduler_thread.daemon=True
        self.scheduler_thread.start()
        logger.debug("Created ComputeDataServiceDecentral")
           
        
    def __get_cds_id(self, cds_url):
        start = cds_url.index(self.CDS_ID_PREFIX)
        end =cds_url.index("/", start)
        return cds_url[start:end]


    ###########################################################################
    # Pilot Compute    
    def add_pilot_compute_service(self, pcs):
        """ Add a PilotComputeService to this CDS.

            @param pcs: The PilotComputeService to which this ComputeDataService will connect.

        """
        self.pilot_job_services.append(pcs)
        CoordinationAdaptor.update_cds(self.url, self)
        if len(self.pilot_job_services)>1:
            logger.error("Decentral ComputeDataService only supports 1 PilotComputeService")
            raise PilotError("Decentral ComputeDataService only supports 1 PilotComputeService")
        

    def remove_pilot_compute_service(self, pcs):
        """ Remove a PilotJobService from this CDS.

            Note that it won't cancel the PilotJobService, it will just no
            longer be connected to this WUS.

            Keyword arguments:
            pilotjob_services -- The PilotJob Service(s) to remove from this
                                 Work Unit Service. 

            Return:
            Result
        """
        self.pilot_job_services.remove(pcs)
        CoordinationAdaptor.update_cds(self.url, self)
        if len(self.pilot_job_services)>1:
            logger.error("Decentral ComputeDataService only supports 1 PilotComputeService")
            raise PilotError("Decentral ComputeDataService only supports 1 PilotComputeService")


    def submit_compute_unit(self, compute_unit_description):
        """ Submit a CU to this Compute Data Service.
            @param compute_unit_description:  The L{ComputeUnitDescription} from the application
            @return: L{ComputeUnit} object
        """
        cu = ComputeUnit(compute_unit_description, self)
        self.compute_units[cu.id]=cu
        self.__submit_cu(cu)        
        return cu
    
    
    def list_pilot_compute(self):
        """ List all pilot compute of CDS """
        return self.pilot_job_service
    
    
    ###########################################################################
    # Compute Data Service private methods
    def __submit_cu(self, compute_unit):
        """ Submits compute unit to Bigjob """
                
        if len(self.pilot_job_services)!=1:
            raise PilotError("No PilotComputeService found. Please start a PCS before submitting ComputeUnits.")
        
        self.__wait_for_du(compute_unit)
        
        sj = subjob()
        self.pcs_coordination_namespace=self.pilot_job_services[0].coordination_queue
        logger.debug("Submit CU to big-job via external queue: %s"%self.pcs_coordination_namespace)
        sj.submit_job(self.pcs_coordination_namespace, compute_unit.subjob_description)
        compute_unit._update_subjob(sj)
        return compute_unit
    
    
    def __wait_for_du(self, compute_unit):
        """ wait for Data Units that are required for Compute Unit """
        cu_description = compute_unit.compute_unit_description
        if cu_description.has_key("input_data") and len(cu_description["input_data"])>0:
            for input_du_url in cu_description["input_data"]:
                for du in self.data_units.values():
                    if input_du_url == du.get_url():
                        logger.debug("Wait for DU: %s"%du.get_url())
                        du.wait()
    
    
    ###########################################################################
    # Pilot Data     
    def add_pilot_data_service(self, pds):
        """ Add a PilotDataService 

            @param pds: The PilotDataService to add.
        """
        self.pilot_data_services.append(pds)
        CoordinationAdaptor.update_cds(self.url, self)
    
    def remove_pilot_data_service(self, pds):
        """ Remove a PilotDataService 
            @param pds: The PilotDataService to remove 
        """
        self.pilot_data_services.remove(pds)
        CoordinationAdaptor.update_cds(self.url, self)
    
     
    def list_pilot_data(self):
        """ List all pilot data of CDS """
        return self.pilot_data_services
    
    
    def list_data_units(self):
        """ List all DUs of CDS """
        return self.data_units.items()
    
    
    def get_data_unit(self, du_id):
        if self.data_units.has_key(du_id):
            return self.data_units[du_id]
        return None
    
    
    def submit_data_unit(self, data_unit_description):
        """ creates a data unit object and binds it to a physical resource (a pilotdata) """
        du = DataUnit(pilot_data=None, 
                      data_unit_description=data_unit_description)
        self.data_units[du.id]=du
        self.du_queue.put(du)
        # queue currently not persisted
        CoordinationAdaptor.update_cds(self.url, self)
        return du
    
    
    ###########################################################################
    # General     
   
    def cancel(self):
        """ Cancel the CDS. 
            All associated PD and PC objects are canceled.            
        """
        # terminate background thread
        self.stop.set()
        CoordinationAdaptor.delete_cds(self.url)
   
   
    def wait(self):
        """ Waits for CUs and DUs. Return after all DU's have been placed (i.e. in state Running)
            and all CU's have been completed (i.e. in state Done) or if a fault has occurred or
            the user has cancelled a CU or DU.            
        """
        try:
            logger.debug("### START WAIT ###")
            for i in self.data_units.values():
                i.wait()
            logger.debug("Wait for DUs finished")        
                
            for i in self.compute_units.values():
                i.wait()     
            logger.debug("CUs done")        
                   
            logger.debug("### END WAIT ###")
        except:
            logger.debug("Ctrl-c detected. Terminating ComputeDataService...")
            self.cancel()
            raise KeyboardInterrupt
                
        
    def get_state(self):
        return self.state
    
    
    def get_id(self):
        return str(self.id)
    
    
    def __del__(self):
        """ Make sure that background thread terminates"""
        self.cancel()
   
   
    ###########################################################################
    # Internal Scheduling
    def __update_scheduler_resources(self):
        logger.debug("__update_scheduler_resources")        
        pd = [s for i in self.pilot_data_services for s in i.list_pilots()]
        self.scheduler.set_pilot_data(pd)
        pj = [p for i in self.pilot_job_services for p in i.list_pilots()]
        logger.debug("Pilot-Jobs: " + str(pj))
        self.scheduler.set_pilot_jobs(pj)
    
    
    def _schedule_du(self, du):
        """ Schedule DU to a suitable pilot data
        
            Currently one level of scheduling is used:
                1.) Add all resources managed by the contained PDS 
                2.) Select one resource
        """ 
        logger.debug("Schedule PD")
        self.__update_scheduler_resources()
        selected_pilot_data = self.scheduler.schedule_pilot_data(du.data_unit_description)
        return selected_pilot_data 
    
    
    def _scheduler_thread(self):
        while True and self.stop.isSet()==False:            
            try:
                #logger.debug("Scheduler Thread: " + str(self.__class__) + " Pilot Data")
                du = self.du_queue.get(True, 1)  
                # check whether this is a real du object  
                if isinstance(du, DataUnit):
                    pd=self._schedule_du(du)                
                    if(pd!=None):                        
                        logger.debug("Initiate Transfer to PD.")
                        du.add_pilot_data(pd)
                        logger.debug("Transfer to PD finished.")
                        du._update_state(State.Running) 
                        self.du_queue.task_done()                   
                    else:
                        self.du_queue.task_done() 
                        self.du_queue.put(du)
            except Queue.Empty:
                pass
               
            if self.du_queue.empty():
                time.sleep(5)        

        logger.debug("Re-Scheduler terminated")
    
    
   
    
   


""" B{PilotData Module}: Implementation of L{PilotData}, L{PilotDataService} and L{DataUnit}
""" 
import sys
import os
import logging
import uuid
import random
import threading
import time
import pdb
import Queue
from pilot.api.api import PilotError

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from bigjob import logger
from pilot.api import PilotData, DataUnit, PilotDataService, State


""" Load file management adaptors """
from pilot.filemanagement.ssh_adaptor import SSHFileAdaptor 
try:
    from pilot.filemanagement.webhdfs_adaptor import WebHDFSFileAdaptor
except:
    logger.warn("WebHDFS package not found.") 
try:
    from pilot.filemanagement.globusonline_adaptor import GlobusOnlineFileAdaptor
except:
    logger.warn("Globus Online package not found.") 
    
try:
    from pilot.filemanagement.gs_adaptor import GSFileAdaptor
except:
    logger.warn("Goggle Storage package not found.") 


try:
    from pilot.filemanagement.s3_adaptor import S3FileAdaptor
except:
    logger.warn("Amazon S3 package not found.") 

try:
    from pilot.filemanagement.irods_adaptor import iRodsFileAdaptor
except:
    logger.warn("iRods Storage package not found.") 



#from pilot.coordination.advert import AdvertCoordinationAdaptor as CoordinationAdaptor
#from pilot.coordination.nocoord import NoCoordinationAdaptor as CoordinationAdaptor
from pilot.coordination.redis_adaptor import RedisCoordinationAdaptor as CoordinationAdaptor
from bliss.saga import Url as SAGAUrl


# generate global application id for this instance
application_id = "bigdata"

"""
    Implementation considerations:
    
    PilotDataService can have n PilotData 
    PilotData can have n DataUnits (only DU URLs is held to avoid circular dependencies)
    A DataUnit can be part of n PilotData 
"""

class PilotData(PilotData):
    """ B{PilotData (PD).} 
                
        This is the object that is returned by the PilotDataService when a 
        new PilotData is created based on a PilotDataDescription. A PilotData represents
        a finite amount of physical space on a certain resource. It can be populated
        with L{DataUnit}s.

        The PilotData object can be used by the application to keep track 
        of a pilot. A PilotData has state, can be queried, can be cancelled.
        
    """   
    
    PD_ID_PREFIX="pd-"   

        
    def __init__(self, pilot_data_service=None, pilot_data_description=None, pd_url=None):    
        """ 
            Initialize PilotData at given service url::
            
                ssh://<hostname>
                gsissh://<hostname>
                go://<hostname>
                gs://google.com
                s3://aws.amazon.com
            
            In the future more SAGA/Bliss URL schemes/adaptors are supported.        
        """ 
        self.id = None
        self.url = pd_url
        self.pilot_data_description = None
        self.pilot_data_service = pilot_data_service
        self.service_url=None
        self.size = None
        self.data_unit_urls = []
        self.security_context = None
        
        if pd_url==None and pilot_data_service!=None:      # new pd          
            self.id = self.PD_ID_PREFIX+str(uuid.uuid1())
            self.pilot_data_description = pilot_data_description
            self.url = CoordinationAdaptor.add_pd(CoordinationAdaptor.get_base_url(application_id)+":"+pilot_data_service.id, self)
        elif pd_url != None:
            logger.warn("Reconnect to PilotData: %s"%pd_url)
            dictionary = CoordinationAdaptor.get_pd(pd_url)
            if dictionary.has_key("security_context"):
                self.security_context=dictionary["security_context"]
            pd_dict = eval(dictionary["pilot_data"])
            for i in pd_dict:
                self.__setattr__(i, pd_dict[i])
            # A Pilot Data does not hold a direct reference to a Data Unit (only URL refs are stored)
            self.data_unit_urls = eval(dictionary["data_unit_urls"])
                        
        self.__initialize_pilot_data()
        CoordinationAdaptor.update_pd(self)
    

    def cancel(self):        
        """ Cancel PilotData  """
        #self.__filemanager.delete_pilotdata()
        pass
    
     
    def get_url(self):
        """ Get URL of PilotData. Used for reconnecting to PilotData """
        return self.url
       
    
    def url_for_du(self, du):
        """ Get full URL to DataUnit within PilotData """
        return self.service_url + "/" + str(du.id)
        

    def submit_data_unit(self, data_unit_description):
        """ creates a data unit object and initially imports data specified in data_unit_description """
        du = DataUnit(pilot_data=self, 
                      data_unit_description=data_unit_description)
        self.data_unit_urls.append(du.get_url())
        du.add_pilot_data(self)
        return du
   
    
    def list_data_units(self):
        """ List all data units of Pilot Data """
        return self.data_unit_urls          
    
    
    def get_state(self):
        """ Return current state of Pilot Data """
        return self.__filemanager.get_state()
    
    
    def get_du(self, du_url):
        """ Returns Data Unit if part of Pilot Data """
        if self.data_unit_urls.count(du_url)>0:
            du = DataUnit(du_url=du_url)
            return du        
        return None
    
    
    def wait(self):
        """ Wait until PD enters a final state (Done, Canceled or Failed).""" 
        while 1:
            finish_counter=0
            result_map = {}
            for du_url in self.data_units_urls: 
                du = DataUnit(du_url=du_url)
                du.wait()
                state = du.get_state()           
                #state = job_detail["state"]                
                if result_map.has_key(state)==False:
                    result_map[state]=1
                else:
                    result_map[state] = result_map[state]+1
                if self.__has_finished(state)==True:
                    finish_counter = finish_counter + 1                   
            logger.debug("PD ID: " + str(self.id) + " Total DUs: %s States: %s"%(len(self.data_units_urls), str(result_map)))
            if finish_counter == len(self.data_units_urls):
                break
            time.sleep(2)

    
    def export_du(self, du, target_url):
        """ Export Data Unit to a local directory """
        if target_url.startswith("/") and os.path.exists(target_url)==False:
            os.mkdir(target_url)
        logger.debug("Export Data-Unit to %s"%target_url)
        self.__filemanager.get_du(du, target_url)
            
                
    def put_du(self, du):
        """Copy Data Unit to Pilot Data"""
        logger.debug("Put DU: %s to Pilot-Data: %s"%(du.id,self.service_url))
        self.__filemanager.create_du(du.id)
        self.__filemanager.put_du(du)
        self.data_unit_urls.append(du.get_url())
        CoordinationAdaptor.update_pd(self)
        
        
    def remove_du(self, du):
        """ Remove Data Unit from Pilot Data """
        if self.data_unit_urls.count(du.get_url())>0:
            self.__filemanager.remove_du(du)
            self.data_unit_urls.remove(du.get_url())
        CoordinationAdaptor.update_pd(self)
        
    
    def copy_du(self, du, pd_new):
        """ Copy DataUnit to another Pilot Data """
        pd_new.create_du(du)
        self.__filemanager.copy_du(du, pd_new)
        
        # update meta data at pd_new
        #pd_new.data_units[du.id] = du
        pd_new.data_unit_urls.append(du.get_url())
        CoordinationAdaptor.update_pd(pd_new)
        
    
    # END API methods
    ###########################################################################
    # Auxillary Methods

    def create_du(self, du):
        """ Create a new Data Unit within Pilot """
        self.__filemanager.create_du(du.id)
  
  
    def __initialize_pilot_data(self):
        
        if self.pilot_data_description!=None:
            self.service_url=self.pilot_data_description["service_url"]
            self.size = self.pilot_data_description["size"]
            
            # initialize file adaptor
            if self.service_url.startswith("ssh:"):
                logger.debug("Use SSH backend")
                self.__filemanager = SSHFileAdaptor(self.service_url,
                                                    self.security_context, 
                                                    self.pilot_data_description)
            elif self.service_url.startswith("http:"):
                logger.debug("Use WebHDFS backend")
                self.__filemanager = WebHDFSFileAdaptor(self.service_url)
            elif self.service_url.startswith("go:"):
                logger.debug("Use Globus Online backend")
                self.__filemanager = GlobusOnlineFileAdaptor(self.service_url)
            elif self.service_url.startswith("gs:"):
                logger.debug("Use Google Cloud Storage backend")
                self.__filemanager = GSFileAdaptor(self.service_url, self.security_context)
            elif self.service_url.startswith("irods:"):
                logger.debug("Use iRods Storage backend")
                self.__filemanager = iRodsFileAdaptor(self.service_url, self.security_context)
            elif self.service_url.startswith("s3:") \
                or self.service_url.startswith("walrus:") \
                or self.service_url.startswith("swift:"):
                logger.debug("Use Amazon S3/Eucalyptus Walrus/SWIFT Storage backend")
                self.__filemanager = S3FileAdaptor(self.service_url, 
                                                   self.security_context, 
                                                   self.pilot_data_description)
            else:
                raise PilotError("No File Plugin found.")
            
            self.__filemanager.initialize_pilotdata()
            self.__filemanager.get_pilotdata_size()
            
            # Update security context
            self.security_context = self.__filemanager.get_security_context()
            

    def __get_pd_id(self, pd_url):
        start = pd_url.index(self.PD_ID_PREFIX)
        end =pd_url.index("/", start)
        return pd_url[start:end]


    
    def to_dict(self):
        """ Internal method that returns a dict with all data contained in this Pilot Data"""
        pd_dict = {}
        pd_dict["id"]=self.id
        pd_dict["url"]=self.url
        pd_dict["pilot_data_description"]=self.pilot_data_description
        logger.debug("PilotData Dictionary: " + str(pd_dict))
        return pd_dict
    
    
    def __repr__(self):
        """Returns Pilot Data URL"""
        return self.service_url
    
    
    def __has_finished(self, state):
        state = state.lower()
        if state=="running" or state=="failed" or state=="canceled":
            return True
        else:
            return False
    
    @classmethod
    def create_pilot_data_from_dict(cls, pd_dict):
        """Restore Pilot Data from dictionary"""
        pd = PilotData()
        for i in pd_dict.keys():
            pd.__setattr__(i, pd_dict[i])
        pd.__initialize_pilot_data()
        logger.debug("created pd " + str(pd))
        return pd
   
    
###############################################################################
COORDINATION_URL = "redis://localhost"

class PilotDataService(PilotDataService):
    """ B{PilotDataService (PDS).}
    
        Factory for creating Pilot Data.
    
    """
    
    PDS_ID_PREFIX="pds-"
    
    def __init__(self, coordination_url=COORDINATION_URL, pds_url=None):
        """ Create a PilotDataService

            Keyword arguments:
            pds_id -- restore from pds_id
        """        
        self.pilot_data={}
        CoordinationAdaptor.configure_base_url(coordination_url)
        if pds_url == None:
            self.id = self.PDS_ID_PREFIX + str(uuid.uuid1())
            application_url = CoordinationAdaptor.get_base_url(application_id)
            self.url = CoordinationAdaptor.add_pds(application_url, self)
        else:
            self.id = self.__get_pds_id(pds_url)
       


    def create_pilot(self, pilot_data_description):
        """ Create a PilotData 

            Keyword arguments:
            pilot_data_description -- PilotData Description:: 
            
                {
                    'service_url': "ssh://<hostname>/base-url/",               
                    'size': "1000"
                }
            
            Return value:
            A PilotData object
        """
        pd = PilotData(pilot_data_service=self, 
                       pilot_data_description=pilot_data_description)
        self.pilot_data[pd.id]=pd
        
        # store pilot data in central data space
        CoordinationAdaptor.add_pd(self.url, pd)        
        return pd
    
    
    def get_pilot(self, pd_id):
        """ Reconnect to an existing pilot. """
        if self.pilot_data.has_key(pd_id):
            return self.pilot_data[pd_id]
        return None


    def list_pilots(self):
        """ List all PDs of PDS """
        return self.pilot_data.values()
    

    def cancel(self):
        """ Cancel the PilotDataService. Release all Pilot Data created by this service.
            
            Keyword arguments:
            None

            Return value:
            Result of operation
        """
        for i in self.pilot_data.values():
            i.cancel()
 
 
    def wait(self):
        """ Wait until all managed PD (of this Pilot Data Service) enter a final state""" 

        for i in self.pilot_data.values():
            i.wait()
 
 
    def get_url(self):
        """ Returns URL of Pilot Data Service """
        return self.url
 
    ###########################################################################
    # Non-API methods
    def to_dict(self):
        """ Return a Python dictionary containing the representation of the PDS 
            (internal method not part of Pilot API)        
        """
        pds_dict = self.__dict__
        pds_dict["id"]=self.id
        return pds_dict
 
 
    def __del__(self):
        """Releases all Pilot Data created by this Pilot Data Service."""
        self.cancel()         
            
    
    def __get_pds_id(self, pds_url):
        start = pds_url.index(self.PDS_ID_PREFIX)
        end =pds_url.index("/", start)
        return pds_url[start:end]

    
    def __restore_pd(self, pds_url):
        pd_list=CoordinationAdaptor.list_pd(pds_url) 
        for i in pd_list:
            pass
        

class DataUnit(DataUnit):
    """ B{DataUnit (DU).}
    
        This is the object that is returned by the ComputeDataService when a 
        new DataUnit is created based on a DataUnitDescription.

        The DataUnit object can be used by the application to keep track 
        of a DataUnit.

        A DataUnit has state, can be queried and can be cancelled.
    
        
    
        State model:
            - New: PD object created
            - Pending: PD object is currently updated  
            - Running: At least 1 replica of PD is persistent in a pilot data            
    """
    
    ## TODO
    # DU are stored as top-level objects in Redis:
    # redis://localhost/<application-id>/du-<id>
    #
    # In the future a DU can be possibly bound to multiple PD
    # Thus, it should be a top level entity
    # The lower levels of the hierarchy will only store references to the DU then
    
    
    DU_ID_PREFIX="du-"  

    def __init__(self, pilot_data=None, data_unit_description=None, du_url=None):
        """
            1.) create a new Pilot Data: pilot_data_service and data_unit_description required
            2.) reconnect to an existing Pilot Data: du_url required 
            
        """
        if du_url==None:
            self.id = self.DU_ID_PREFIX + str(uuid.uuid1())
            self.data_unit_description = data_unit_description        
            self.pilot_data=[]
            self.state = State.New
            self.data_unit_items=[]
            if self.data_unit_description.has_key("file_urls"):
                self.data_unit_items = DataUnitItem.create_data_unit_list(self, self.data_unit_description["file_urls"]) 

            self.url = None

            # register a data unit as top-level entry in Redis
            application_url = CoordinationAdaptor.get_base_url(application_id)
            self.url = CoordinationAdaptor.add_du(application_url, self)
            CoordinationAdaptor.update_du(self)
            
            # Deprecated
            # old method only allowed the creation of a du if a pd existed
            #if pilot_data!=None:
            #    # Allow data units that are not connected to a resource!
            #    self.url = CoordinationAdaptor.add_du(pilot_data.url, self)
            #    CoordinationAdaptor.update_du(self)
        else:
            self.id = DataUnit._get_du_id(du_url)
            self.url = du_url   
            logger.debug("Restore du: %s"%self.id)         
            self.__restore_state()
            
        self.transfer_threads=[]
               
            
    def cancel(self):
        """ Cancel the Data Unit. """
        self.state = State.Done    
        if len(self.pilot_data) > 0: 
            CoordinationAdaptor.update_du(self)

            
    def add_files(self, file_url_list=[]):
        """Add files referenced in list to Data Unit"""
        self.state=State.Pending
        item_list = DataUnitItem.create_data_unit_from_urls(None, file_url_list)
        for i in item_list:
            self.data_unit_items.append(i)
        CoordinationAdaptor.update_du(self)    
        if len(self.pilot_data) > 0: 
            for i in self.pilot_data:
                logger.debug("Update Pilot Data %s"%(i.get_url()))
                i.put_du(self)
        CoordinationAdaptor.update_du(self)    
        
        
    def remove_files(self, file_urls):
        """Remove files from Data Unit (NOT implemented yet"""
        # TODO
        #self.data_unit_items.remove(input_data_unit)
        if len(self.pilot_data) > 0:
            CoordinationAdaptor.update_du(self)

        
    def list(self):
        """ List all items contained in DU 
            {
                "filename" : { 
                                "pilot_data" : [url1, url2],
                                "local" : url
                             }
            }        
        """        
        self.__refresh()
        base_urls = [i.url_for_du(self) for i in self.get_pilot_data()]
        result_dict = {}
        for i in self.data_unit_items:
            logger.debug("Process file: %s"%(i.filename))
            result_dict[i.filename]={
                                    "pilot_data": [os.path.join(j, i.filename) for j in base_urls],
                                    "local": i.local_url
                                    }
        return result_dict
    
   
    
    def get_state(self):
        """ Return current state of DataUnit """
        # update remote state
        du_dict = CoordinationAdaptor.get_du(self.url)
        self.state = du_dict["state"]
        return self.state  
    
    
    def wait(self):
        """ Wait until in running state 
            (or failed state)
        """
        logger.debug("DU: %s wait()"%(str(self.id)))
        # Wait for all transfers to finish
        for i in self.transfer_threads:
            i.join()
        
        # Wait for state to change
        while self.state!=State.Running and self.get_state()!=State.Failed:
            logger.debug("Waiting DU State: %s"%self.state)
            time.sleep(2)
    
    
    def add_pilot_data(self, pilot_data):
        """ add this DU (self) to a certain pilot data 
            data will be moved into this data
        """
        transfer_thread=threading.Thread(target=self.__add_pilot_data, args=[pilot_data])
        transfer_thread.start()        
        self.transfer_threads.append(transfer_thread)
        
    
    def get_pilot_data(self):
        """ get a list of pilot data that have a copy of this PD """
        self.__restore_state()
        return self.pilot_data
    
    
    def export(self, target_url):
        """ simple implementation of export: 
                copies file from first pilot data to local machine
        """
        if self.state!=State.Running:
            self.wait()
        if len(self.pilot_data) > 0:
            self.pilot_data[0].export_du(self, target_url)
        else:
            logger.error("No Pilot Data for PD found")
    
    
    def get_url(self):
        """ Return URL that can be used to reconnect to Data Unit """
        return self.url
    
    
    
    ###########################################################################
    # BigData Internal Methods
    def to_dict(self):
        """ Internal method that returns a dict with all data contained in this DataUnit"""
        du_dict = self.__dict__
        du_dict["id"]=self.id
        return du_dict        
    
    
    def _update_state(self, state):
        """ Internal method for updating state"""
        self.state=state
        logger.debug("Update DU state to " + state + " Number of PD: " + str(len(self.pilot_data)))
        if len(self.pilot_data) > 0: 
            CoordinationAdaptor.update_du(self)

    
    def __add_pilot_data(self, pilot_data):
        logger.debug("add du to pilot data")
        if len(self.pilot_data) > 0: # copy files from other pilot data
            self.pilot_data[0].copy_du(self, pilot_data)
        else: # copy files from original location
            pilot_data.put_du(self)
        self.pilot_data.append(pilot_data)
        self.state = State.Running
        
        #self.url = CoordinationAdaptor.add_du(pilot_data.url, self)
        CoordinationAdaptor.update_du(self)
            
        
    @classmethod    
    def _get_du_id(cls, du_url):
        try:
            start = du_url.index(cls.DU_ID_PREFIX)
            end = du_url.find("/", start)
            if end==-1:
                end = du_url.find("?", start)
            if end==-1:
                end = len(du_url)
            return du_url[start:end]
        except:
            logger.error("No valid PD URL")
        return None
    
    
    def __refresh(self):
        """ Update list of data units items 
            from coordination service """
        try:
            if self.url != None:
                du_dict = CoordinationAdaptor.get_du(self.url)
                data_unit_dict_list = eval(du_dict["data_unit_items"])
                self.data_unit_items = [DataUnitItem.create_data_unit_from_dict(i) for i in data_unit_dict_list]
        except:
            logger.warn("Refresh of DU %s failed"%(self.get_url()))
            
        
    def __restore_state(self):
        du_dict = CoordinationAdaptor.get_du(self.url)
        # Restore Data Unit
        self.data_unit_description = eval(du_dict["data_unit_description"])
        self.state = du_dict["state"]
        
        # Restore DataUnitItems
        data_unit_dict_list = eval(du_dict["data_unit_items"])
        self.data_unit_items = [DataUnitItem.create_data_unit_from_dict(i) for i in data_unit_dict_list]
        
        # restore Pilot Data
        pd_list = eval(du_dict["pilot_data"])
        self.pilot_data = [] 
        for i in pd_list:
            logger.debug("PD: "+str(i)) 
            pd = PilotData(pd_url=str(i))
            self.pilot_data.append(pd) 


    def __repr__(self):        
        return "PD: " + str(self.url) 
        + " \nData Units: " + str(self.data_unit_items)
        + " \nPilot Data: " + str(self.pilot_data)
    
    

class DataUnitItem(object):
    """ DataUnitItem """
    DUI_ID_PREFIX="dui-"  
   
    def __init__(self, pd=None, local_url=None):        
        if local_url!=None:
            self.id = self.DUI_ID_PREFIX + str(uuid.uuid1())
            self.local_url = local_url   
            self.filename =  os.path.basename(local_url)    
            #if pd != None:
            #    self.url = pd.url + "/" + self.filename
        
        
    @classmethod    
    def __exists_file(cls, url):   
        """ return True if file at url exists. Otherwise False """
        file_url = SAGAUrl(url)
        if file_url.host == "":
            if os.path.exists(str(file_url)):
                return True
            else:
                return False            
        elif file_url.host=="localhost":
            if os.path.exists(file_url.path):
                return True
            else:
                return False
        else:            
            return True
        
    
    def __repr__(self):
        return str(self.__dict__) 
        
        
    ###########################################################################
    # Auxiliary Methods
    @classmethod
    def create_data_unit_list(cls, pd=None, urls=None):
        """ Creates a list of DUs from URL list
        """    
        du_list = []    
        for i in urls:            
            if cls.__exists_file(i):
                du = DataUnitItem(pd, i)
                du_list.append(du)
    
        return du_list
    
    @classmethod
    def create_data_unit_from_urls(cls, pd=None, urls=None):
        """ Creates a list of DUs from URL list
        """    
        du_item_list = []    
        for i in urls:            
            if cls.__exists_file(i):
                du = DataUnitItem(pd, i)
                du_item_list.append(du)
    
        return du_item_list
    
    
    @classmethod
    def create_data_unit_from_dict(cls, du_dict):
        du = DataUnitItem()
        logger.debug("Restore DU: " + str(du_dict))
        for i in du_dict.keys():
            logger.debug("Set attribute: %s", i)
            du.__setattr__(i, du_dict[i])
        return du
    
    
    def to_dict(self):
        du_dict = self.__dict__
        du_dict["id"]=self.id
        return du_dict
    
###################################################################################################    
# Tests
# Auxilliary testing methods
def __get_pd_url(du_url):
    url = du_url[:du_url.index(":du-")]
    return url

def __get_du_id(du_url):
    du_id = du_url[du_url.index("du-"):]
    return du_id

# Tests
def test_pd_reconnect():
    du_url = "redis://localhost/bigdata:pds-f31a670c-e3f6-11e1-afaf-705681b3df0f:pd-f31c47b8-e3f6-11e1-af44-705681b3df0f:du-f4debce8-e3f6-11e1-8399-705681b3df0f"
    pd_url = __get_pd_url(du_url)
    pd = PilotData(pd_url=pd_url)
    print str(pd.list_data_units())
    du = pd.get_du(du_url)
    
    #du = DataUnit(du_url="redis://localhost/bigdata:pds-32d63b2e-df05-11e1-a329-705681b3df0f:pd-37674138-df05-11e1-80d0-705681b3df0f:du-3b8d428c-df05-11e1-af2a-705681b3df0f")
    logger.debug(str(du.list()))


def test_du_reconnect():
    du_url = "redis://localhost/bigdata:du-1d1b7078-229f-11e2-834e-705681b3df0f"
    du = DataUnit(du_url=du_url)
    logger.debug(str(du.list()))
    du.export("/tmp/export-test")
    
    
def test_data_unit_add_file():
    pilot_data_service = PilotDataService(coordination_url="redis://localhost/")
    pilot_data_description = {
                                "service_url": "ssh://localhost/tmp/pilot-" + str(uuid.uuid1()),
                                "size": 100                                   
                             }
    pd = pilot_data_service.create_pilot(pilot_data_description=pilot_data_description)
    
    # create data unit for output data
    output_data_unit_description = {
         "file_urls": [], 
         "file_url_patterns": ["test.txt"]                             
    }
    output_data_unit = pd.submit_data_unit(output_data_unit_description)
    output_data_unit.wait()
    logger.debug("Output DU: " + output_data_unit.get_url())
    pd_reconnect_url = pd.get_url()
    du_url = output_data_unit.get_url()
    pd_reconnect = PilotData(pd_url=pd_reconnect_url)
    du_reconnect = pd_reconnect.get_du(du_url)
    du_reconnect.add_files(["test.txt"])
    
    

if __name__ == "__main__":
    #test_data_unit_add_file()
    test_du_reconnect()
    



import sys
import os
import time
import logging
import uuid
#logging.basicConfig(level=logging.DEBUG)

#sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot import PilotComputeService, PilotDataService, ComputeDataService, State
from bigjob import logger 

#COORDINATION_URL = "redis://localhost:6379"
COORDINATION_URL="redis://ILikeBigJob_wITH-REdIS@gw68.quarry.iu.teragrid.org:6379/pcs/pcs-4867ff08-e192-11e1-a694-00003e980000"

if __name__ == "__main__":      
    
    print COORDINATION_URL
    # create pilot data service (factory for data pilots (physical, distributed storage))
    # and pilot data
    pilot_data_service = PilotDataService(coordination_url=COORDINATION_URL)
    
    ###################################################################################################
    # Pick one of the Pilot Data Descriptions below    
    
    pilot_data_description_aws={
                                "service_url": "s3://pilot-data-" + str(uuid.uuid1()),
                                "size": 100,   
                                "region" : "us-west-1", # or "" for DEFAULT/EAST
                                "access_key_id":"AKIAJPGNDJRYIG5LIEUA",
                                "secret_access_key":"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"                                                       
                              }
    
    pilot_data_description_india = {
                                "service_url": "walrus://149.165.146.135/pilot-data-" + str(uuid.uuid1()),
                                #"service_url": "ssh://localhost/tmp/pilot-data-" + str(uuid.uuid1()),
                                "affinity_datacenter_label": "us-east",              
                                "affinity_machine_label": "",
                                "access_key_id":"",
                                "secret_access_key":""
                             }
    
    
    pd = pilot_data_service.create_pilot(pilot_data_description=pilot_data_description_aws)
     
     
    # Create Data Unit Description
    #base_dir = "../data1"
    #url_list = os.listdir(base_dir)
    # make absolute paths
    #absolute_url_list = [os.path.join(base_dir, i) for i in url_list]
    data_unit_description = {
                              "file_urls": [os.path.join(os.getcwd(), "test.txt")],
                              "affinity_datacenter_label": "us-east-1",              
                              "affinity_machine_label": ""
                             }    
      
    # submit pilot data to a pilot store 
    input_data_unit = pd.submit_data_unit(data_unit_description)
    input_data_unit.wait()
    
    logger.info("Data Unit URL: " + input_data_unit.get_url())
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)
    
    pilot_compute_description_amazon_west = {
                             "service_url": 'ec2+ssh://aws.amazon.com',
                             "number_of_processes": 1,                             
                             'affinity_datacenter_label': "us-google",              
                             'affinity_machine_label': "", 
                             # cloud specific attributes
                             #"vm_id":"ami-d7f742be",
                             "vm_id": "ami-5c3b1b19",
                             "vm_ssh_username":"ubuntu",
                             "vm_ssh_keyname":"MyKey",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/id_rsa",
                             "vm_type":"t1.micro",
                             "region" : "us-west-1",
                             "access_key_id":"AKIAJPGNDJRYIG5LIEUA",
                             "secret_access_key":"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"
                           
                            }
    
    
    pilot_compute_description_euca_india = {
                             "service_url": 'euca+ssh://149.165.146.135:8773/services/Eucalyptus',
                             #"service_url": 'fork://localhost',
                             "number_of_processes": 1,
                             'affinity_datacenter_label': "us-east",              
                             'affinity_machine_label': "", 
                             #'working_directory': os.getcwd(),
                             # cloud specific attributes
                             "vm_id":"emi-36913A82",
                             "vm_ssh_username":"root",
                             "vm_ssh_keyname":"luckow",
                             "vm_ssh_keyfile":"/Users/luckow/.ssh/eucakey-india",
                             "vm_type":"c1.xlarge",
                             "access_key_id":"",
                             "secret_access_key":""
                            }
    
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description_amazon_west)
    
    compute_data_service = ComputeDataService()
    compute_data_service.add_pilot_compute_service(pilot_compute_service)
    compute_data_service.add_pilot_data_service(pilot_data_service)
    
    # create empty data unit for output data
    output_data_unit_description = {
         "file_urls": []              
    }
    output_data_unit = pd.submit_data_unit(output_data_unit_description)
    output_data_unit.wait()
    
    # create compute unit
    compute_unit_description = {
            "executable": "/bin/cat",
            "arguments": ["test.txt"],
            "number_of_processes": 1,
            "output": "stdout.txt",
            "error": "stderr.txt",   
            "input_data": [input_data_unit.get_url()],
            # Put files stdout.txt and stderr.txt into output data unit
            "output_data": [
                            {
                             output_data_unit.get_url(): 
                             ["std*"]
                            }
                           ]    
    }   
    
    compute_unit = compute_data_service.submit_compute_unit(compute_unit_description)
    logger.info("Finished setup of ComputeDataService. Waiting for scheduling of PD")
    compute_data_service.wait()
    
    logger.debug("Output Data Unit: " + str(output_data_unit.list()))
    
    logger.info("Terminate Pilot Compute/Data Service")
    compute_data_service.cancel()
    pilot_data_service.cancel()
    pilot_compute_service.cancel()



""" B{BigJob-based Implementation of Pilot-API} 

This package contains the implementations of the Pilot API objects: L{PilotComputeService}, L{PilotCompute}, L{PilotDataService}, L{PilotData}, L{ComputeDataService}.

"""


import sys
import os
import time
import logging
logging.basicConfig(level=logging.DEBUG)

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot import PilotComputeService, ComputeDataService, State

COORDINATION_URL = "redis://localhost:6379"

if __name__ == "__main__":      
    
    pilot_compute_service = PilotComputeService(coordination_url=COORDINATION_URL)

    # create pilot job service and initiate a pilot job
    pilot_compute_description = {
                             "service_url": 'fork://localhost',
                             "number_of_processes": 1,                             
                             "working_directory": os.path.join(os.getcwd(), "agent"),
                             "affinity_datacenter_label": "eu-de-south",              
                             "affinity_machine_label": "mymachine-1",
                             "file_transfer": ["ssh://" + os.path.dirname(os.path.abspath(__file__)) 
                                               + "/../test.txt > BIGJOB_WORK_DIR"]
                            }
    
    pilotjob = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description)
    
    # start compute unit
    compute_unit_description = {
            "executable": "/bin/cat",
            "arguments": ["test.txt"],
            "number_of_processes": 1,
            "output": "stdout.txt",
            "error": "stderr.txt",   
            "affinity_datacenter_label": "eu-de-south",              
            "affinity_machine_label": "mymachine-1",
            "file_transfer": ["ssh://" + os.path.dirname(os.path.abspath(__file__)) 
                                + "/../test.txt > BIGJOB_WORK_DIR"]
    }    
   
    compute_unit = pilotjob.submit_compute_unit(compute_unit_description)
    logging.debug("Finished submission. Waiting for completion of CU")
    compute_unit.wait()
    
    
    logging.debug("Terminate Pilot Compute Service")
    pilot_compute_service.cancel()


'''
iRods based File Transfer Implementation
'''
import urlparse

import errno
import sys
import os
import stat
import logging
import traceback
import time
import re
import shutil
import pdb
import glob
import pexpect

# This is for local debugging!
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "../.."))

import bliss.saga as saga
from pilot.api import State
from bigjob import logger



class iRodsFileAdaptor(object):
    """ BigData File Management for Pilot Data 
        Supports pilot data on top of iRods.
        
            Assumption: working iRods installation
        
            irods://localhost/${OSG_DATA}/?vo=<vo>&resource-group=<resource-group>
            
    """ 
    def __init__(self, resource_url, security_context=None, pilot_data_description=None):        
        self.resource_url = saga.Url(resource_url)
        query_string = self.resource_url.query
        self.localpath = self.resource_url.path
        self.vo = re.search("(?<=vo=)(.*)([&\b]{1})", query_string).group(1)
        self.resource_group = re.search("(?<=resource-group=)(.*)[&\b$]?", query_string).group(1)
        logger.debug("VO: %s, Resource Group: %s"%(self.vo, self.resource_group))
        self.is_local = self.__is_local()
        
    
    def __is_local(self):
        # test whether path contains environment variable
        match = re.search("\$\{(.*)\}", self.localpath)
        if match:
            env_var = match.group(1)
            logger.debug("Found: " + env_var + " in URL.")
            logger.debug("Env list: " + str(os.environ))
            if os.environ.has_key(env_var):
                self.localpath = re.sub(r'\$\{.*\}', os.environ[env_var], self.localpath)
                #self.localpath = os.environ[env_var]
                logger.debug("Expanding URL Path to: " + self.localpath)
                return True
        logger.debug("No expansion in: " + self.localpath)
        return False
    
    
    def get_security_context(self):
        """ Returns security context that needs to be available on the distributed
            node in order to access this Pilot Data """
        return None
                       
        
    def initialize_pilotdata(self):
        pass
       
        
    def get_pilotdata_size(self):
        # unlimited size
        return None
    
    
    def delete_pilotdata(self):
        self.__state=State.Done
    
        
    def get_state(self):
        return self.__state
        
            
    def create_du(self, du_id):
        logger.debug("create iRods collection: " + du_id)
        if self.is_local:
            command = "mkdir %s"%(os.path.join(self.localpath, du_id))
        else:
            command = "imkdir %s"%(du_id)
        self.__run_command(command)

                 
    def put_du(self, du):
        start = time.time()
        logger.debug("Copy DU to iRod")
        du_items = du.list()
        for i in du_items.keys():     
            try:
                local_filename=du_items[i]["local"]
                remote_path = os.path.join(str(du.id), os.path.basename(local_filename))
                logger.debug("copy %s to %s"%(local_filename, remote_path))
                self._put_file(local_filename, remote_path)
            except:
                logger.debug("Could not copy: " + str(i))
        logger.debug("Finished Put DU in: " + str(time.time()-start) + " sec.")
                
    
    def get_du(self, du, target_url):
        #du_id = "du-7370d7b5-ed0b-11e1-95df-705681b3df0f"
        start = time.time()
        du_id = du.id
        logger.debug("Get DU: " + str(du_id))
        if self.is_local:
            command = "cp -r %s %s"%(os.path.join(self.localpath, du_id), target_url)
            source_path = os.path.join(self.localpath, du_id, "*")
            target_path = target_url
            logger.debug("Target and source host are localhost. Processing: %s" %(source_path))
            expanded_path = glob.glob(source_path)
            logger.debug("Expanded path: " + str(expanded_path))
            for path in expanded_path:
                if os.path.isdir(path):
                    logger.debug("Source path %s is directory"%path)
                    files = os.listdir(path)
                    for i in files:
                        try:
                            os.symlink(os.path.join(files, i), target_path)
                            os.chmod(os.path.join(target_path, os.path.basename(path)), 0777)
                        except:
                            self.__print_traceback()
                else:
                    try:
                        os.symlink(path, os.path.join(target_path, os.path.basename(path)))
                    except:
                        self.__print_traceback()

        else:
            command = "iget -f -r %s %s"%(du_id, target_url)
            logger.debug(command)
            self.__run_command(command)
            
            full_path = os.path.join(target_url, du_id)
            #logger.debug("Path: " + str(full_path) + " Exists: " + str(os.path.exists(full_path)))
            #while os.path.exists(full_path)==False:
            #    time.sleep(1)
    
            for i in os.listdir(full_path):
                try:
                    logger.debug("chmod " + str(i))
                    os.chmod(os.path.join(full_path, i), 0777)
                    logger.debug("move " + str(i))
                    shutil.move(os.path.join(full_path, i), target_url)
                except:
                    self.__print_traceback()
    
            shutil.rmtree(full_path, ignore_errors=True)
            #time.sleep(2)
            #if target_url==".":
            #    target_url = os.getcwd()
            #command = "mv %s/* %s"%(os.path.join(target_url, du_id), target_url)
            #self.__run_command(command)
            logger.debug("Finished Get DU " + du.id + " in: " + str(time.time()-start) + " sec.")
        
   
    def copy_du(self, du, pd_new):
        remote_url = pd_new.resource_url + "/" + str(du.id)
        local_url =  self.resource_url  + "/" + str(du.id)
        self.copy_du_to_url(du, local_url, remote_url)  
            
        
    def remove_du(self, du):
        if self.is_local:
            command = "rm -rf %s"%(os.path.join(self.localpath, du.id))
        else:
            command = "irm %s"%du.id
        
        self.__run_command(command)
    
    
    ###########################################################################
    # Pure File Management APIs
    def _put_file(self, source, target):
        logger.debug("Put file: %s to %s"%(source, target))
        start = time.time()
        if self.is_local:
            command = "cp -r %s %s"%(source, target)
        else:
            command = "iput -f -R %s %s %s"%(self.resource_group, source, target)
        self.__run_command(command)
        put_time = time.time() - start
        number_replica = 0
        if self.is_local==False:
            #pdb.set_trace()
            home_directory= self.__run_command("ipwd")[0].strip()
            full_filename = os.path.join(home_directory, target)
            command = "irepl-osg -f %s -G %s"%(full_filename, self.resource_group)
            output = self.__run_command(command) 
            for i in output:
                if i.find("copied") > 0 or i.find("replica")>0:
                    number_replica = number_replica + 1 
        rep_time = time.time() - start - put_time
        logger.info("Upload;Replication;Total;File Size;Backend;Number Replica: %f;%f;%f;%d;%s;%d"%(put_time, rep_time, time.time()-start, os.path.getsize(source), self.resource_group, number_replica))
         

    def transfer(self, source_url, target_url):
        pass
    
    
    def create_remote_directory(self, target_url):
        return True
    
                   
    ###########################################################################
    def __run_command(self, command):
        logger.debug(command)
        child = pexpect.spawn(command, timeout=None)
        output = child.readlines()
        logger.debug("Run %s Output: %s"%(command, str(output)))
        child.close()
        return output 

    def __print_traceback(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        print "*** print_tb:"
        traceback.print_tb(exc_traceback, limit=1, file=sys.stderr)
        print "*** print_exception:"
        traceback.print_exception(exc_type, exc_value, exc_traceback,
                              limit=2, file=sys.stderr)
    
    
def test_irods():
    irods = iRodsFileAdaptor("irods://gw68/${OSG_DATA}/osg/home/luckow/?vo=osg&resource-group=osgGridFtpGroup")
    irods.initialize_pilotdata()
    irods.create_du("du-7370d7b5-ed0b-11e1-95df-705681b3df0f")
    irods._put_file("test.txt", "du-7370d7b5-ed0b-11e1-95df-705681b3df0f/test.txt")
    irods.get_du("du-7370d7b5-ed0b-11e1-95df-705681b3df0f", "export")



if __name__ == "__main__":
    test_irods()
    



import compute.api
import data.api

class PilotError(Exception):
    def __init__(self, value):
        self.value = value
    
    def __str__(self):
        return repr(self.value)


class ComputeDataService(compute.api.ComputeUnitService, data.api.PilotDataService):
    """ B{ComputeDataService (CDS).}
    
        The ComputeDataService is the application's interface to submit 
        ComputeUnits and PilotData/DataUnit to the Pilot-Manager 
        in the P* Model.        
    """
   

    def __init__(self, wds_id=None):
        """ Create a Compute Data Service object.

            Keyword arguments:
            wds_id -- Reconnect to an existing WDS (optional).
        """
        raise  NotImplementedError("Abstract super class, please use ComputeDataService implementation class in pilot namespace")
   




import sys
import os
import time
import logging
import json

logging.basicConfig(level=logging.DEBUG)

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))
from pilot import PilotDataService, ComputeDataService, DataUnit, State

COORDINATION_URL = "redis://localhost:6379"

if __name__ == "__main__":        
    
    if len(sys.argv)==2:
        reconnect_url=sys.argv[1]
    else:
        print "Usage: " + sys.executable + " " + __file__ + " <Data Unit URL to Reconnect to>"
        sys.exit(-1)
        
    # create pilot data service (factory for pilot stores (physical, distributed storage))
    pilot_data_service = PilotDataService(coordination_url=COORDINATION_URL)
    pd_new = pilot_data_service.create_pilot({
                                'service_url': "ssh://localhost/tmp/pilotdata-reconnect/",
                                'size':100,
                               'affinity_datacenter_label': "eu-de-south",              
                               'affinity_machine_label': "mymachine-1"
                                })
    
    
    logging.debug("Pilot Data URL: %s"%pilot_data_service.url)
    
    
    ###########################################################################
    # PD should only be scheduled to machine 1    
    logging.debug("Connect to PD URL: %s"%reconnect_url)
    pd = DataUnit(du_url=reconnect_url)
    
    
    # Move PD to another pilot store    
    pd.add_pilot_data(pd_new)
    
    #time.sleep(120)
    #pilot_data_service.cancel()


import sys, os
import stat
import urlparse

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)) + "/../../../webhdfs-py/")

from pilot.api import State
from bigjob import logger
logger.debug(str(sys.path))

from webhdfs.webhdfs import WebHDFS 


class WebHDFSFileAdaptor(object):
    
    HDFS_USER_NAME="luckow"
    HDFS_SERVICE_HOST="192.168.2.108"
    HDFS_SERVICE_PORT=50070     
    
    def __init__(self, service_url):     
        
        self.service_url = service_url
        
        try:
            result = urlparse.urlparse(service_url)
            self.host = result.netloc
            self.path = result.path        
        except:
            logger.error("Error parsing URL.")
            
        self.__state=State.New
        self.__webhdfs= WebHDFS(self.HDFS_SERVICE_HOST, 
                               self.HDFS_SERVICE_PORT,
                               self.HDFS_USER_NAME)
        
    def get_security_context(self):
        """ Returns security context that needs to be available on the distributed
            node in order to access this Pilot Data """
        return None
        
        
    def initialize_pilotstore(self):
        self.__webhdfs.mkdir(self.path)
        
        
    def get_pilotstore_size(self):
        return 0
    
    
    def delete_pilotstore(self):
        self.__webhdfs.rmdir(self.path)
        
    def get_state(self):
        return self.__state
            
            
    def create_pd(self, pd_id):
        pd_dir = self.__get_pd_path(pd_id)
        logger.debug("mkdir: " + pd_dir) 
        self.__webhdfs.mkdir(pd_dir)
        
        
    def put_pd(self, pd):
        for i in pd.list_data_units():     
            remote_path = os.path.join(self.__get_pd_path(pd.id), os.path.basename(i.local_url))
            logger.debug("Put file: %s to %s"%(i.local_url, remote_path))
                        
            if i.local_url.startswith("file://") or i.local_url.startswith("/"):
                if stat.S_ISDIR(os.stat(i.local_url).st_mode):
                    logger.warning("Path %s is a directory. Ignored."%i.local_url)                
                    continue            
                self.__webhdfs.copyFromLocal(i.local_url, remote_path)
            else:
                logger.error("File URLs: %s not supported"%i.local_url)
                             

    def copy_pd_to_url(self, pd,  local_url, remote_url):
        
        if not remote_url.startswith("file://") and not remote_url.startswith("/"):
            logger.error("Only local URLs supported")
            return
        
        result = urlparse.urlparse(remote_url)
        path = result.path    
        # create directory
        try:
            os.makedirs(path)
        except:
            logger.debug("Directory: %s already exists."%path)
            
        base_dir = self.__get_pd_path(pd.id)
        for filename in self.__webhdfs.listdir(base_dir):
            file_url = local_url + "/" + filename
            file_remote_url = remote_url + "/" + filename
            logger.debug("GET " + file_url + " to " + file_remote_url)
            self.__webhdfs.copyToLocal(file_url, file_remote_url)



    def copy_pd(self, pd, ps_new):
        pass
    

    def get_pd(self, pd, target_url):
        remote_url = target_url
        local_url =  self.__get_pd_path(pd.id)
        self.copy_pd_to_url(pd, local_url, remote_url)         
    
        
    def remove_pd(self, pd):
        self.__webhdfs.rmdir(self.__get_pd_path(pd.id))
    
    
    ###########################################################################
    # Internal methods
    def __get_pd_path(self, pd_id):
        return os.path.join(self.path, str(pd_id))
    


'''
Command Line Util for using BigJob (via the Pilot-API)
'''
import argparse
import sys
import os
import pdb
import pickle

import bigjob
from pilot import PilotComputeService, PilotCompute, ComputeUnit, State
BIGJOB_DIRECTORY="~/.bigjob/" 
  
  
class BigJobCLI(object):
    
    def __init__(self):
        self.pilots = []
        self.__restore()
        
    def submit_pilot(self, coordination_url="redis://localhost/",
                           resource_url="fork://localhost",
                           working_directory=".",
                           number_cores=1,
                           cores_per_node=1,
                     ):
        
        # create pilot job service and initiate a pilot job
        pilot_compute_description = {
                             "service_url": resource_url,
                             "number_of_processes": 1,                             
                             "working_directory": os.path.join(os.getcwd(), working_directory)  ,
                             "number_of_processes": number_cores,
                             "processes_per_node": cores_per_node                             
                            }
        self.submit_pilot_by_description(coordination_url, pilot_compute_description)
        
    
    def submit_pilot_by_description(self, coordination_url="redis://localhost/", pilot_compute_description={}):
        pilot_compute_service = PilotComputeService(coordination_url=coordination_url)
        pilot_compute = pilot_compute_service.create_pilot(pilot_compute_description=pilot_compute_description)
        pilot_url = pilot_compute.get_url()
        self.pilots.append(pilot_url)
        print("Started Pilot: %s"%(pilot_url))
        self.__persist()
        

    def cancel_pilot(self, pilot_url):
        pilot_compute = PilotCompute(pilot_url=pilot_url)
        pilot_compute.cancel()
        self.pilots.remove(pilot_url)
        self.__persist()
        
    
    def list_pilots(self):
        print "\nPilot Compute\t\t\t\t\t\t\t\t\tState"
        print "-----------------------------------------------------------------------------------------------------"
        if len(self.pilots)==0:
            print "No pilot found"
        for i in self.pilots:
            pilot_compute = PilotCompute(pilot_url=i)
            print "%s\t%s"%(pilot_compute.get_url(), pilot_compute.get_state())
        print ""
        
    
    def list_cus(self, pilot_url):
        pilot_compute = PilotCompute(pilot_url=pilot_url)
        cus = pilot_compute.list_compute_units()
        counter=1
        print "\nPilot Compute: %s"%(pilot_compute.get_url())
        print "State: %s"%(pilot_compute.get_state())
        print "#\tCompute Unit\t\t\t\t\tState\tQueue\tRuntime"
        print "-----------------------------------------------------------------------------------------"
        if len(cus)==0:
            print "No Compute Unit found."
        for i in cus:
            url = i.get_url()
            short_url = url[url.index("sj"):]
            if short_url.find("/?")!=-1:
                short_url=short_url[:short_url.index("/?")]
            queue_time = 0.0
            run_time = 0.0
            try:
                pass
                #details = i.get_details()
                #state = details["state"]
            except:
                pass
                
            state = i.get_state()
            details=i.get_details()
            if details.has_key("start_time") and details.has_key("end_queue_time"):
                queue_time = float(details["end_queue_time"]) - float(details["start_time"])
            if details.has_key("end_time") and details.has_key("end_queue_time"):
                run_time = float(details["end_time"]) - float(details["end_queue_time"])
            print "%d\t%s\t\t%s\t%.1f\t%.1f"%(counter, short_url, state, queue_time, run_time)
            counter = counter + 1 
        print ""
            

    def submit_cu(self, pilot_url, command):
        """ submits CUs (does not waits for completion) """
        #print "Submit CU to %s"%(pilot_url)
        args= []
        if len(command)>1:
            args = command[1:]
        compute_unit_description = {
            "executable": command[0],
            "arguments": args,
            "total_core_count": 1,
            "number_of_processes": 1,            
            "output": "stdout.txt",
            "error": "stderr.txt",
        }    
        return self.submit_cu_by_description(pilot_url, compute_unit_description)
    
    
    
    def submit_cu_by_description(self, pilot_url, compute_unit_description={}):
        pilot_compute = PilotCompute(pilot_url=pilot_url)
        compute_unit = pilot_compute.submit_compute_unit(compute_unit_description)
        print "Started ComputeUnit: %s"%(compute_unit.get_url())
        return compute_unit
            
        
    def run_cu(self, pilot_url, command):
        """ submits CU and waits for completion """
        compute_unit=self.submit_cu(pilot_url, command)
        print "Waiting for termination"
        compute_unit.wait()
        print "CU %s terminated"%compute_unit.get_url()
        return compute_unit
    
    
    def run_cu_by_description(self, pilot_url, compute_unit_description={}):
        compute_unit=self.submit_cu_by_description(pilot_url, compute_unit_description)
        print "Waiting for termination"
        compute_unit.wait()
        print "CU %s terminated"%compute_unit.get_url()
        return compute_unit
    
    
    def wait_cu(self, pilot_url):
        pilot_compute = PilotCompute(pilot_url=pilot_url)
        pilot_compute.wait()
        
        
    def cancel_cu(self, cu_url):
        compute_unit = ComputeUnit(cu_url=cu_url)
        compute_unit.cancel()
        print("Terminated CU: %s"%(cu_url))
    
    
    def get_cu_state(self, cu_url):
        compute_unit = ComputeUnit(cu_url=cu_url)
        print "Compute Unit: %s State: %s"%(cu_url, compute_unit.get_state())
    
    ###########################################################################
    # auxiliary methods

    def version(self):
        print "BigJob Version: " + bigjob.version
    
    def clean(self):
        os.remove(self.__get_save_filename())
    
    ###########################################################################
    # private and protected methods
    
    def __persist(self):
        f = open(self.__get_save_filename(), 'wb')
        pickle.dump(self.pilots, f)
        f.close()
    
    def __restore(self):
        if os.path.exists(self.__get_save_filename()):
            try:
                f = open(self.__get_save_filename(), 'rb')
                self.pilots = pickle.load(f)
                f.close()
            except:
                pass

    def __get_save_filename(self):
        return os.path.join(os.path.expanduser(BIGJOB_DIRECTORY), 'pilot-cli.p')


def main():
    app = BigJobCLI()
    parser = argparse.ArgumentParser(add_help=True, description="""BigJob Command Line Utility""")
    
    parser.add_argument('--coordination', '-c', default="redis://localhost")
    parser.add_argument('--clean', action="store_true", default=False)
    parser.add_argument('--version', action="store_true", default=False)    
    
    pilot_group = parser.add_argument_group('Manage pilots')
    pilot_group.add_argument('--number_cores', default="1")
    pilot_group.add_argument('--cores_per_node',  default="1")    
    pilot_group.add_argument('--submit_pilot', action="store", nargs="?", metavar="RESOURCE_URL", 
                              help="submit a pilot to specified resource, e.g. fork://localhost",
                              default=False)    
    pilot_group.add_argument('--pilot_description', action="store", metavar="PILOT_COMPUTE_DESCRIPTION", 
                           help="""accepts a Pilot-API pilot compute description in JSON format: 
                                   \"{'service_url': 'fork://localhost', 'number_of_processes': 1, 'working_directory': '/tmp'}\"""", 
                                default=False)
    
    
    pilot_group.add_argument('--cancel_pilot', action="store", default=False, metavar="PILOT_URL", 
                               help="Cancel pilot")
    pilot_group.add_argument('--list_pilots', action="store_true", default=False, help="list all pilots")
    pilot_group.add_argument('--wait_cus', action="store_true", default=False, help="wait for termination of all CUs")
    pilot_group.add_argument('--list_cus', action="store",  metavar="PILOT_URL", default=False)
    
    
    cu_group = parser.add_argument_group('Manage compute units')
    cu_group.add_argument('--submit_cu', action="store", nargs="+", metavar=("PILOT_URL", "COMMAND ARGS"), 
                           help="submit CU to pilot", default=False)
    cu_group.add_argument('--compute_unit_description', action="store", metavar="COMPUTE_UNIT_DESCRIPTION", 
                           help="""accepts a Pilot-API compute unit description in JSON format: 
                           \"{'executable': '/bin/date', 'arguments': [''], 'total_core_count': 1, 'number_of_processes': 1, 'output': 'stdout.txt', 'error': 'stderr.txt'}\"""", 
                            default=False)
    
    cu_group.add_argument('--run_cu', action="store", nargs="+", metavar=("PILOT_URL", "COMMAND ARGS"), 
                           help="submit CU to pilot and wait for completion", default=False)
    cu_group.add_argument('--get_cu_state', action="store", metavar="CU_URL", default=False)
    cu_group.add_argument('--cancel_cu', action="store", metavar="CU_URL", default=False)
    
    parsed_arguments = parser.parse_args()    
    #print(str(parsed_arguments))
    
    if parsed_arguments.submit_pilot!=False:
        if parsed_arguments.pilot_description!=False:
            pilot_description = eval(parsed_arguments.pilot_description)
            app.submit_pilot_by_description(coordination_url=parsed_arguments.coordination,
                                            pilot_compute_description=pilot_description)            
        else:
            app.submit_pilot(coordination_url=parsed_arguments.coordination,
                             resource_url=parsed_arguments.submit_pilot,
                             number_cores=parsed_arguments.number_cores,
                             cores_per_node=parsed_arguments.cores_per_node)    
    elif parsed_arguments.list_pilots!=False:
        app.list_pilots()
    elif parsed_arguments.cancel_pilot!=False:
        app.cancel_pilot(parsed_arguments.cancel_pilot)
    elif parsed_arguments.submit_cu!=False:
        if parsed_arguments.compute_unit_description!=False:
            compute_unit_description = eval(parsed_arguments.compute_unit_description)
            app.submit_cu_by_description(parsed_arguments.submit_cu[0], 
                                         compute_unit_description=compute_unit_description)            
        else:
            app.submit_cu(parsed_arguments.submit_cu[0], parsed_arguments.submit_cu[1:])
    elif parsed_arguments.run_cu!=False:
        if parsed_arguments.compute_unit_description!=False:
            compute_unit_description = eval(parsed_arguments.compute_unit_description)
            app.run_cu_by_description(parsed_arguments.run_cu[0],
                                            compute_unit_description=compute_unit_description)            
        else:
            app.run_cu(parsed_arguments.run_cu[0], parsed_arguments.run_cu[1:])    
    elif parsed_arguments.get_cu_state!=False:
        app.get_cu_state(parsed_arguments.get_cu_state)
    elif parsed_arguments.wait_cus!=False:
        app.wait_cu(parsed_arguments.wait_cus)
    elif parsed_arguments.list_cus!=False:
        app.list_cus(parsed_arguments.list_cus)
    elif parsed_arguments.cancel_cu!=False:
        app.list_cus(parsed_arguments.cancel_cu)
    elif parsed_arguments.clean==True:
        app.clean()
    elif parsed_arguments.version==True:
        app.version()
        
if __name__ == '__main__':
    main()
    
    


#!/usr/bin/env python
"""Dynamic BigJob (ManyJob) with affinity aware scheduler.
"""

import os
import sys
sys.path.append(os.path.dirname( __file__ ))
from many_job import *

class many_job_affinity_service(many_job_service):
    
    def __init__(self, bigjob_list, advert_host):
        """ accepts resource list as key/value pair:
            ( {"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_cores" : "32", "allocation" : "<allocaton>", "queue" : "workq", 
               "re_agent":  os.getcwd() + "/bigjob_agent_launcher.sh"  "walltime":1000, "affinity": "affinity1"},
              {"resource_url" : "gram://qb1.loni.org/jobmanager-pbs", "number_cores" : "32", "allocation" : "<allocation", "queue" : "workq", 
               "re_agent": "os.getcwd() + "/bigjob_agent_launcher.sh", "walltime":1000, "affinity": "affinity1"})
        """   
        super(many_job_affinity_service, self).__init__(bigjob_list, advert_host)
        
    def __schedule_subjob (self, subjob):
        """ find resource (bigjob) with the same affinity for subjob 
            returns bigjob object """
        for i in self.bigjob_list:
            bigjob = i["bigjob"]
            lock = i["lock"]
            affinity = i["affinity"]
            lock.acquire()
            free_cores = i["free_cores"]
            bigjob_url = bigjob.pilot_url
            state = bigjob.get_state_detail()
            env  = subjob.job_description.environment[0]
            sj_affinity=None
            if env.find("affinity") >=0:
                sj_affinity = env.split("=")[1]            
            logging.debug("Subjob Env: " + str(sj_affinity))
            logging.debug("Big Job: " + bigjob_url + " Cores: " + "%s"%free_cores + "/" + i["number_cores"] + " Affinity: " 
                          + affinity + " SJ affinity: " + sj_affinity + " State: " + state)
            if state.lower() == "running" and free_cores >= int(subjob.job_description.number_of_processes) and affinity == sj_affinity:
                free_cores = i["free_cores"]
                free_cores = free_cores - int(subjob.job_description.number_of_processes)
                i["free_cores"]=free_cores
                lock.release()
                return i 

            lock.release()

        # no resource found
        self.subjob_queue.put(subjob)
        logging.debug("found no active resource for sub-job => (re-) queue it")
        return None        
    
    


""" Abstract API Definition of the B{Pilot-API}; This is not the implementation! Please see L{pilot.impl}"""

from pilot.api.compute.api import PilotCompute
from pilot.api.compute.api import PilotComputeService
from pilot.api.compute.api import PilotComputeDescription
from pilot.api.compute.api import ComputeUnit
from pilot.api.compute.api import ComputeUnitService
from pilot.api.compute.api import ComputeUnitDescription
from pilot.api.compute.api import State

from pilot.api.data.api import PilotDataDescription
from pilot.api.data.api import PilotData
from pilot.api.data.api import PilotDataService
from pilot.api.data.api import DataUnitService
from pilot.api.data.api import DataUnit
from pilot.api.data.api import DataUnitDescription

from pilot.api.api import ComputeDataService
from pilot.api.api import PilotError
